# cython: language_level=3
# cython: profile=True
# Time-stamp: <2020-11-24 17:37:21 Tao Liu>
"""Module Description: Build shifting model

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# Python modules
# ------------------------------------
import sys, time, random
import array

# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.Utilities.Constants import *
from MACS3.Signal.Pileup import naive_quick_pileup, naive_call_peaks

# ------------------------------------
# Other modules
# ------------------------------------
from cpython cimport bool
from cpython cimport array
import numpy as np
cimport numpy as np

# ------------------------------------
# C lib
# ------------------------------------
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t

ctypedef np.float32_t float32_t

class NotEnoughPairsException(Exception):
    def __init__ (self,value):
        self.value = value
    def __str__ (self):
        return repr(self.value)

cdef class PeakModel:
    """Peak Model class.
    """
    cdef:
        object treatment
        double gz
        int max_pairnum
        int umfold
        int lmfold
        int bw
        int d_min
        int tag_expansion_size
        object info, debug, warn, error
        str summary
        int max_tags
        int peaksize
        public np.ndarray plus_line, minus_line, shifted_line
        public int d
        public int scan_window
        public int min_tags
        public list alternative_d
        public np.ndarray xcorr, ycorr

    def __init__ ( self, opt , treatment, int max_pairnum=500 ): #, double gz = 0, int umfold=30, int lmfold=10, int bw=200, int ts = 25, int bg=0, bool quiet=False):
        self.treatment = treatment
        self.gz = opt.gsize
        self.umfold = opt.umfold
        self.lmfold = opt.lmfold
        self.tag_expansion_size = 10         #opt.tsize| test 10bps. The reason is that we want the best 'lag' between left & right cutting sides. A tag will be expanded to 10bps centered at cutting point.
        self.d_min = opt.d_min               #discard any predicted fragment sizes < d_min
        self.bw = opt.bw
        self.info  = opt.info
        self.debug = opt.debug
        self.warn  = opt.warn
        self.error = opt.warn
        self.max_pairnum = max_pairnum

    cpdef build (self):
        """Build the model. Main function of PeakModel class.

        1. prepare self.d, self.scan_window, self.plus_line,
        self.minus_line and self.shifted_line.

        2. find paired + and - strand peaks

        3. find the best d using x-correlation
        """
        cdef:
            dict paired_peakpos
            long num_paired_peakpos
            bytes c                       #chromosome

        self.peaksize = 2*self.bw
        self.min_tags = int(round(float(self.treatment.total) * self.lmfold * self.peaksize / self.gz / 2)) # mininum unique hits on single strand, decided by lmfold
        self.max_tags = int(round(float(self.treatment.total) * self.umfold * self.peaksize / self.gz /2)) # maximum unique hits on single strand, decided by umfold
        self.debug( f"#2 min_tags: {self.min_tags}; max_tags:{self.max_tags}; " )
        
        self.info( "#2 looking for paired plus/minus strand peaks..." )
        # find paired + and - strand peaks
        paired_peakpos = self.__find_paired_peaks ()

        num_paired_peakpos = 0
        for c in list( paired_peakpos.keys() ):
            num_paired_peakpos += len (paired_peakpos[c] )

        self.info("#2 Total number of paired peaks: %d" % (num_paired_peakpos))

        if num_paired_peakpos < 100:
            self.error(f"#2 MACS3 needs at least 100 paired peaks at + and - strand to build the model, but can only find {num_paired_peakpos}! Please make your MFOLD range broader and try again. If MACS3 still can't build the model, we suggest to use --nomodel and --extsize 147 or other fixed number instead.")
            self.error("#2 Process for pairing-model is terminated!")
            raise NotEnoughPairsException("No enough pairs to build model")

        # build model, find the best d using cross-correlation
        self.__paired_peak_model(paired_peakpos)

    def __str__ (self):
        """For debug...

        """
        return """
Summary of Peak Model:
  Baseline: %d
  Upperline: %d
  Fragment size: %d
  Scan window size: %d
""" % (self.min_tags,self.max_tags,self.d,self.scan_window)

    cdef dict __find_paired_peaks (self):
        """Call paired peaks from fwtrackI object.

        Return paired peaks center positions.
        """
        cdef:
           int i
           list chrs
           bytes chrom
           np.ndarray[np.int32_t, ndim=1] plus_tags, minus_tags
           list plus_peaksinfo
           list minus_peaksinfo
           dict paired_peaks_pos # return

        chrs = list(self.treatment.get_chr_names())
        chrs.sort()
        paired_peaks_pos = {}
        for i in range( len(chrs) ):
            chrom = chrs[ i ]
            self.debug( f"Chromosome: {chrom}" )
            # extract tag positions
            [ plus_tags, minus_tags ] = self.treatment.get_locations_by_chr( chrom )
            # look for + strand peaks
            plus_peaksinfo = self.__naive_find_peaks ( plus_tags )
            self.debug("Number of unique tags on + strand: %d" % ( plus_tags.shape[0] ) )
            self.debug("Number of peaks in + strand: %d" % ( len(plus_peaksinfo) ) )
            if plus_peaksinfo:
                self.debug(f"plus peaks: first - {plus_peaksinfo[0]} ... last - {plus_peaksinfo[-1]}")
            # look for - strand peaks
            minus_peaksinfo = self.__naive_find_peaks ( minus_tags )
            self.debug("Number of unique tags on - strand: %d" % ( minus_tags.shape[0] ) )
            self.debug("Number of peaks in - strand: %d" % ( len( minus_peaksinfo ) ) )
            if minus_peaksinfo:
                self.debug(f"minus peaks: first - {minus_peaksinfo[0]} ... last - {minus_peaksinfo[-1]}")
            if not plus_peaksinfo or not minus_peaksinfo:
                self.debug("Chrom %s is discarded!" % (chrom) )
                continue
            else:
                paired_peaks_pos[chrom] = self.__find_pair_center (plus_peaksinfo, minus_peaksinfo)
                self.debug("Number of paired peaks in this chromosome: %d" %(len(paired_peaks_pos[chrom])))
        return paired_peaks_pos

    cdef list __naive_find_peaks ( self, np.ndarray[np.int32_t, ndim=1] taglist ):
        """Naively call peaks based on tags counting.

        Return peak positions and the tag number in peak region by a tuple list [(pos,num)].
        """
        cdef:
            long i
            int pos
            list peak_info
            list pileup_array

        peak_info = []    # store peak pos in every peak region and
                          # unique tag number in every peak region
        if taglist.shape[0] < 2: # less than 2 tags, no need to call peaks, return []
            return peak_info

        pileup_array = naive_quick_pileup( taglist, self.peaksize/2 ) # build pileup by extending both side to half peak size
        peak_info = naive_call_peaks( pileup_array, self.min_tags, self.max_tags )

        return peak_info

    cdef __paired_peak_model (self, dict paired_peakpos,):
        """Use paired peak positions and treatment tag positions to build the model.

        Modify self.(d, model_shift size and scan_window size. and extra, plus_line, minus_line and shifted_line for plotting).
        """
        cdef:
            int window_size, i
            list chroms
            object paired_peakpos_chrom
            np.ndarray[np.int32_t, ndim=1] tags_plus, tags_minus, plus_start, plus_end, minus_start, minus_end, plus_line, minus_line
            np.ndarray plus_data, minus_data, xcorr, ycorr, i_l_max

        window_size = 1+2*self.peaksize+self.tag_expansion_size
        self.plus_line = np.zeros(window_size, dtype="int32") # for plus strand pileup
        self.minus_line = np.zeros(window_size, dtype="int32")# for minus strand pileup
        plus_start = np.zeros(window_size, dtype="int32")     # for fast pileup
        plus_end = np.zeros(window_size, dtype="int32")       # for fast pileup
        minus_start = np.zeros(window_size, dtype="int32")    # for fast pileup
        minus_end = np.zeros(window_size, dtype="int32")      # for fast pileup
        self.debug("start model_add_line...")
        chroms = list(paired_peakpos.keys())

        for i in range(len(chroms)):
            paired_peakpos_chrom = paired_peakpos[chroms[i]]
            (tags_plus, tags_minus) = self.treatment.get_locations_by_chr(chroms[i])
            # every paired peak has plus line and minus line
            self.__model_add_line (paired_peakpos_chrom, tags_plus, plus_start, plus_end) #, plus_strand=1)
            self.__model_add_line (paired_peakpos_chrom, tags_minus, minus_start, minus_end) #, plus_strand=0)

        self.__count ( plus_start, plus_end, self.plus_line )
        self.__count ( minus_start, minus_end, self.minus_line )

        self.debug("start X-correlation...")
        # Now I use cross-correlation to find the best d
        plus_line = self.plus_line
        minus_line = self.minus_line

        # normalize first
        minus_data = (minus_line - minus_line.mean())/(minus_line.std()*len(minus_line))
        plus_data = (plus_line - plus_line.mean())/(plus_line.std()*len(plus_line))

        # cross-correlation
        ycorr = np.correlate(minus_data,plus_data,mode="full")[window_size-self.peaksize:window_size+self.peaksize]
        xcorr = np.linspace(len(ycorr)//2*-1, len(ycorr)//2, num=len(ycorr))

        # smooth correlation values to get rid of local maximums from small fluctuations.
        ycorr = smooth(ycorr, window="flat") # window size is by default 11.

        # all local maximums could be alternative ds.
        i_l_max = np.r_[False, ycorr[1:] > ycorr[:-1]] & np.r_[ycorr[:-1] > ycorr[1:], False]
        i_l_max = np.where(i_l_max)[0]
        i_l_max = i_l_max[ xcorr[i_l_max] > self.d_min ]
        i_l_max = i_l_max[ np.argsort(ycorr[i_l_max])[::-1]]

        self.alternative_d = sorted([int(x) for x in xcorr[i_l_max]])
        assert len(self.alternative_d) > 0, "No proper d can be found! Tweak --mfold?"

        self.d = xcorr[i_l_max[0]]

        self.ycorr = ycorr
        self.xcorr = xcorr

        self.scan_window = max(self.d,self.tag_expansion_size)*2

        self.info("#2 Model building with cross-correlation: Done")

        return True

    cdef __model_add_line (self, list pos1, np.ndarray[np.int32_t, ndim=1] pos2, np.ndarray[np.int32_t, ndim=1] start, np.ndarray[np.int32_t, ndim=1] end): #, int plus_strand=1):
        """Project each pos in pos2 which is included in
        [pos1-self.peaksize,pos1+self.peaksize] to the line.

        pos1: paired centers -- list of coordinates
        pos2: tags of certain strand -- a numpy.array object
        line: numpy array object where we pileup tags

        """
        cdef:
            int i1, i2, i2_prev, i1_max, i2_max, last_p2, psize_adjusted1, psize_adjusted2, p1, p2, max_index, s, e

        i1 = 0                  # index for pos1
        i2 = 0                  # index for pos2
        i2_prev = 0             # index for pos2 in previous pos1
                                # [pos1-self.peaksize,pos1+self.peaksize]
                                # region
        i1_max = len(pos1)
        i2_max = pos2.shape[0]
        last_p2 = -1
        flag_find_overlap = False

        max_index = start.shape[0] - 1

        psize_adjusted1 = self.peaksize + self.tag_expansion_size // 2 # half window

        while i1<i1_max and i2<i2_max:
            p1 = pos1[i1]
            p2 = pos2[i2] #- self.tag_expansion_size/2

            if p1-psize_adjusted1 > p2: # move pos2
                i2 += 1
            elif p1+psize_adjusted1 < p2: # move pos1
                i1 += 1
                i2 = i2_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    i2_prev = i2 # only the first index is recorded
                # project
                s = max(int(p2-self.tag_expansion_size/2-p1+psize_adjusted1), 0)
                start[s] += 1
                e = min(int(p2+self.tag_expansion_size/2-p1+psize_adjusted1), max_index)
                end[e] -= 1
                i2+=1
        return

    cdef __count ( self, np.ndarray[np.int32_t, ndim=1] start, np.ndarray[np.int32_t, ndim=1] end, np.ndarray[np.int32_t, ndim=1] line ):
        """
        """
        cdef:
            int i
            long pileup
        pileup = 0
        for i in range(line.shape[0]):
            pileup += start[i] + end[i]
            line[i] = pileup
        return

    cdef __find_pair_center (self, list pluspeaks, list minuspeaks):
        cdef:
            long ip = 0                  # index for plus peaks
            long im = 0                  # index for minus peaks
            long im_prev = 0             # index for minus peaks in previous plus peak
            list pair_centers
            long ip_max
            long im_max
            bool flag_find_overlap
            int pp, mp
            float pn, mn

        pair_centers = []
        ip_max = len(pluspeaks)
        im_max = len(minuspeaks)
        self.debug(f"ip_max: {ip_max}; im_max: {im_max}")
        flag_find_overlap = False
        while ip<ip_max and im<im_max:
            (pp,pn) = pluspeaks[ip] # for (peakposition, tagnumber in peak)
            (mp,mn) = minuspeaks[im]
            if pp-self.peaksize > mp: # move minus
                im += 1
            elif pp+self.peaksize < mp: # move plus
                ip += 1
                im = im_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    im_prev = im # only the first index is recorded
                if pn/mn < 2 and pn/mn > 0.5: # number tags in plus and minus peak region are comparable...
                    if pp < mp:
                        pair_centers.append((pp+mp)//2)
                        #self.debug ( "distance: %d, minus: %d, plus: %d" % (mp-pp,mp,pp))
                im += 1
        if pair_centers:
            self.debug(f"Paired centers: first - {pair_centers[0]} ... second - {pair_centers[-1]} ")
        return pair_centers

# smooth function from SciPy cookbook: http://www.scipy.org/Cookbook/SignalSmooth
cpdef smooth(x, int window_len=11, str window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError, "Input vector needs to be bigger than window size."


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"


    s=np.r_[x[window_len-1:0:-1],x,x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='valid')
    return y[(window_len//2):-(window_len//2)]




# cython: language_level=3
# cython: profile=True
# Time-stamp: <2022-04-14 17:47:31 Tao Liu>

"""Module Description: statistics functions to calculate p-values

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from math import fabs
from math import log1p #as py_log1p
from math import sqrt
import sys
# ------------------------------------
# Other modules
# ------------------------------------
import numpy as np
cimport numpy as np
from numpy cimport uint8_t, uint16_t, uint32_t, uint64_t, int8_t, int16_t, int32_t, int64_t, float32_t, float64_t
from cpython cimport bool

# ------------------------------------
# C lib
# ------------------------------------
from libc.math cimport exp,log,log10, M_LN10 #,fabs,log1p

# ------------------------------------
# constants
# ------------------------------------
cdef int32_t LSTEP = 200
cdef float64_t EXPTHRES = exp(LSTEP)
cdef float64_t EXPSTEP  = exp(-1*LSTEP)
cdef float64_t bigx = 20

# ------------------------------------
# Normal distribution functions
# ------------------------------------
# x is the value, u is the mean, v is the variance
cpdef pnorm(int32_t x, int32_t u, int32_t v):
    """The probability of X=x when X=Norm(u,v)
    """
    return 1.0/sqrt(2.0 * 3.141592653589793 * <float32_t>(v)) * exp(-<float32_t>(x-u)**2 / (2.0 * <float32_t>(v)))

cpdef float32_t pnorm2(float32_t x, float32_t u, float32_t v):
    """The probability of X=x when X=Norm(u,v)
    """
    cdef:
        float32_t ret

    try:
        ret =1.0/sqrt(2.0 * 3.141592653589793 * v) * exp(-(x-u)**2 / (2.0 * v))
    except ValueError:
        sys.exit(1)
    return ret

# ------------------------------------
# Misc functions
# ------------------------------------

cpdef factorial ( uint32_t n ):
    """Calculate N!.

    """
    cdef float64_t fact = 1
    cdef uint64_t i
    if n < 0:
        return 0
    for i in range( 2,n+1 ):
        fact = fact * i
    return fact

cdef float64_t poz(float64_t z):
    """ probability of normal z value
    """
    cdef:
        float64_t y, x, w
        float64_t Z_MAX = 6.0 # Maximum meaningful z value
    if z == 0.0:
        x = 0.0;
    else:
        y = 0.5 * fabs(z)
        if y >= (Z_MAX * 0.5):
            x = 1.0
        elif y < 1.0:
            w = y * y
            x = ((((((((0.000124818987 * w
                        - 0.001075204047) * w + 0.005198775019) * w
                        - 0.019198292004) * w + 0.059054035642) * w
                        - 0.151968751364) * w + 0.319152932694) * w
                        - 0.531923007300) * w + 0.797884560593) * y * 2.0
        else:
            y -= 2.0
            x = (((((((((((((-0.000045255659 * y
                             + 0.000152529290) * y - 0.000019538132) * y
                             - 0.000676904986) * y + 0.001390604284) * y
                             - 0.000794620820) * y - 0.002034254874) * y
                             + 0.006549791214) * y - 0.010557625006) * y
                             + 0.011630447319) * y - 0.009279453341) * y
                             + 0.005353579108) * y - 0.002141268741) * y
                             + 0.000535310849) * y + 0.999936657524
    if z > 0.0:
        return ((x + 1.0) * 0.5)
    else:
        return ((1.0 - x) * 0.5)

cdef float64_t ex20 ( float64_t x ):
    """Wrapper on exp function. It will return 0 if x is smaller than -20.
    """
    if x < -20.0:
        return 0.0
    else:
        return exp(x)

cpdef float64_t chisq_pvalue_e ( float64_t x, uint32_t df ):
    """Chisq CDF function for upper tail and even degree of freedom.
    Good for p-value calculation and designed for combining pvalues.

    Note: we assume df to be an even number larger than 1! Do not
    violate this assumption and there is no checking.

    df has to be even number! if df is odd, result will be wrong!

    """
    cdef:
        float64_t  a, y, s
        float64_t  e, c, z

    if x <= 0.0:
        return 1.0

    a = 0.5 * x
    even = ((2*(df/2)) == df)
    y = ex20(-a)
    s = y
    if df > 2:
        x = 0.5 * (df - 1.0)
        z = 1.0
        if a > bigx:            # approximation for big number
            e = 0.0
            c = log (a)
            while z <= x:
                e = log (z) + e
                s += ex20(c*z-a-e)
                z += 1.0
            return s
        else:
            e = 1.0
            c = 0.0
            while z <= x:
                e = e * (a / z)
                c = c + e
                z += 1.0
            return c * y + s
    else:
        return s

cpdef float64_t chisq_logp_e ( float64_t x, uint32_t df, bool log10 = False ):
    """Chisq CDF function in log space for upper tail and even degree of freedom
    Good for p-value calculation and designed for combining pvalues.

    Note: we assume df to be an even number larger than 1! Do not
    violate this assumption and there is no checking.

    Return value is -logp. If log10 is set as True, return -log10p
    instead.

    """
    cdef:
        float64_t a, y
        float64_t s                # s is the return value
        float64_t e, c, z

    if x <= 0.0:
        return 0.0

    a = 0.5 * x
    y = exp(-a)             # y is for small number calculation
    # initialize s
    s = -a
    if df > 2:
        x = 0.5 * (df - 1.0)    # control number of iterations
        z = 1.0             # variable for iteration
        if a > bigx:            # approximation for big number
            e = 0.0         # constant
            c = log (a)         # constant
            while z <= x:       # iterations
                e += log(z)
                s = logspace_add(s, c*z-a-e)
                z += 1.0
        else:                   # for small number
            e = 1.0             # not a constant
            c = 0.0             # not a constant
            while z <= x:
                e = e * (a / z)
                c = c + e
                z += 1.0
            s = log(y+c*y) #logspace_add( s, log(c) ) - a
    # return
    if log10:
        return -s/log(10)
    else:
        return -s

cpdef float64_t poisson_cdf ( uint32_t n, float64_t lam, bool lower=False, bool log10=False ):
    """Poisson CDF evaluater.

    This is a more stable CDF function. It can tolerate large lambda
    value. While the lambda is larger than 700, the function will be a
    little slower.

    Parameters:
    n     : your observation
    lam   : lambda of poisson distribution
    lower : if lower is False, calculate the upper tail CDF, otherwise, to calculate lower tail; Default is False.
    log10 : if log10 is True, calculation will be in log space. Default is False.
    """
    assert lam > 0.0, "Lambda must > 0, however we got %d" % lam

    if log10:
        if lower:
            # lower tail
            return log10_poisson_cdf_P_large_lambda(n, lam)
        else:
            # upper tail
            return log10_poisson_cdf_Q_large_lambda(n, lam)

    if lower:
        if lam > 700: # may be problematic
            return __poisson_cdf_large_lambda (n, lam)
        else:
            return __poisson_cdf(n,lam)
    else:
        # upper tail
        if lam > 700: # may be problematic
            return __poisson_cdf_Q_large_lambda (n, lam)
        else:
            return __poisson_cdf_Q(n,lam)

cdef inline float64_t __poisson_cdf ( uint32_t k, float64_t a ):
    """Poisson CDF For small lambda. If a > 745, this will return
    incorrect result.

    Parameters:
    k	: observation
    a	: lambda
    """
    cdef:
        float64_t nextcdf
        float64_t cdf
        uint32_t i
        float64_t lastcdf

    if k < 0:
        return 0.0                        # special cases

    nextcdf = exp( -1 * a )
    cdf = nextcdf

    for i in range( 1, k + 1 ):
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i
        cdf = cdf + nextcdf
    if cdf > 1.0:
        return 1.0
    else:
        return cdf

cdef inline float64_t __poisson_cdf_large_lambda ( uint32_t k, float64_t a ):
    """Slower poisson cdf for large lambda ( > 700 )

    Parameters:
    k	: observation
    a	: lambda
    """
    cdef:
        int32_t num_parts
        float64_t lastexp
        float64_t nextcdf
        float64_t cdf
        uint32_t i
        float64_t lastcdf

    assert a > 700
    if k < 0:
        return 0.0                        # special cases

    num_parts = <int32_t>( a / LSTEP )
    lastexp = exp( -1 * ( a % LSTEP ) )
    nextcdf = EXPSTEP

    num_parts -= 1

    for i in range( 1 , k + 1 ):
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i
        cdf = cdf + nextcdf
        if nextcdf > EXPTHRES or cdf > EXPTHRES:
           if num_parts>=1:
               cdf *= EXPSTEP
               nextcdf *= EXPSTEP
               num_parts -= 1
           else:
               cdf *= lastexp
               lastexp = 1

    for i in range( num_parts ):
        cdf *= EXPSTEP
    cdf *= lastexp
    return cdf

cdef inline float64_t __poisson_cdf_Q ( uint32_t k, float64_t a ):
    """internal Poisson CDF evaluater for upper tail with small
    lambda.

    Parameters:
    k	: observation
    a	: lambda
    """
    cdef uint32_t i

    if k < 0:
        return 1.0                        # special cases
    cdef float64_t nextcdf
    nextcdf = exp( -1 * a )
    cdef float64_t lastcdf

    for i in range(1,k+1):
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i

    cdef float64_t cdf = 0.0
    i = k+1
    while nextcdf >0.0:
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i
        cdf += nextcdf
        i+=1
    return cdf

cdef inline float64_t __poisson_cdf_Q_large_lambda ( uint32_t k, float64_t a ):
    """Slower internal Poisson CDF evaluater for upper tail with large
    lambda.

    Parameters:
    k	: observation
    a	: lambda
    """
    assert a > 700
    if k < 0: return 1.0                        # special cases
    cdef uint32_t num_parts = <int32_t>(a/LSTEP)
    cdef float64_t lastexp = exp(-1 * (a % LSTEP) )
    cdef float64_t nextcdf = EXPSTEP
    cdef uint32_t i
    cdef float64_t lastcdf

    num_parts -= 1

    for i in range(1,k+1):
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i
        if nextcdf > EXPTHRES:
           if num_parts>=1:
               nextcdf *= EXPSTEP
               num_parts -= 1
           else:
               # simply raise an error
               raise Exception("Unexpected error")
               #cdf *= lastexp
               #lastexp = 1
    cdef float64_t cdf = 0.0
    i = k+1
    while nextcdf >0.0:
        lastcdf = nextcdf
        nextcdf = lastcdf * a / i
        cdf += nextcdf
        i+=1
        if nextcdf > EXPTHRES or cdf > EXPTHRES:
           if num_parts>=1:
               cdf *= EXPSTEP
               nextcdf *= EXPSTEP
               num_parts -= 1
           else:
               cdf *= lastexp
               lastexp = 1

    for i in range(num_parts):
        cdf *= EXPSTEP
    cdf *= lastexp
    return cdf

cdef inline float64_t log10_poisson_cdf_P_large_lambda ( uint32_t k, float64_t lbd ):
    """Slower Poisson CDF evaluater for lower tail which allow
    calculation in log space. Better for the pvalue < 10^-310.

    Parameters:
    k	: observation
    lbd	: lambda

    ret = -lambda + \ln( \sum_{i=k+1}^{\inf} {lambda^i/i!} = -lambda + \ln( sum{ exp{ln(F)} } ), where F=lambda^m/m!
    \ln{F(m)} = m*ln{lambda} - \sum_{x=1}^{m}\ln(x)
    Calculate \ln( sum{exp{N} ) by logspace_add function

    Return the log10(pvalue)
    """
    cdef float64_t residue = 0
    cdef float64_t logx = 0
    cdef float64_t ln_lbd = log(lbd)

    # first residue
    cdef int32_t m = k
    cdef float64_t sum_ln_m = 0
    cdef int32_t i = 0
    for i in range(1,m+1):
        sum_ln_m += log(i)
    logx = m*ln_lbd - sum_ln_m
    residue = logx

    while m > 1:
        m -= 1
        logy = logx-ln_lbd+log(m)
        pre_residue = residue
        residue = logspace_add(pre_residue,logy)
        if fabs(pre_residue-residue) < 1e-10:
            break
        logx = logy

    return round((residue-lbd)/M_LN10,5)

cdef inline float64_t log10_poisson_cdf_Q_large_lambda ( uint32_t k, float64_t lbd ):
    """Slower Poisson CDF evaluater for upper tail which allow
    calculation in log space. Better for the pvalue < 10^-310.

    Parameters:
    k	: observation
    lbd	: lambda

    ret = -lambda + \ln( \sum_{i=k+1}^{\inf} {lambda^i/i!} = -lambda + \ln( sum{ exp{ln(F)} } ), where F=lambda^m/m!
    \ln{F(m)} = m*ln{lambda} - \sum_{x=1}^{m}\ln(x)
    Calculate \ln( sum{exp{N} ) by logspace_add function

    Return the log10(pvalue)
    """
    cdef float64_t residue = 0
    cdef float64_t logx = 0
    cdef float64_t ln_lbd = log(lbd)

    # first residue
    cdef int32_t m = k+1
    cdef float64_t sum_ln_m = 0
    cdef int32_t i = 0
    for i in range(1,m+1):
        sum_ln_m += log(i)
    logx = m*ln_lbd - sum_ln_m
    residue = logx

    while True:
        m += 1
        logy = logx+ln_lbd-log(m)
        pre_residue = residue
        residue = logspace_add(pre_residue,logy)
        if fabs(pre_residue-residue) < 1e-5:
            break
        logx = logy

    return round((residue-lbd)/log(10),5)

cdef inline float64_t logspace_add ( float64_t logx, float64_t logy ):
    """addition in log space.

    Given two log values, such as logx and logy, return
    log(exp(logx)+exp(logy)).

    """
    if logx > logy:
        return logx + log1p( exp ( logy - logx ) )
    else:
        return logy + log1p( exp ( logx - logy ) )

cpdef poisson_cdf_inv ( float64_t cdf, float64_t lam, int32_t maximum=1000 ):
    """inverse poisson distribution.

    cdf : the CDF
    lam : the lambda of poisson distribution

    note: maxmimum return value is 1000
    and lambda must be smaller than 740.
    """
    assert lam < 740
    if cdf < 0 or cdf > 1:
        raise Exception ("CDF must >= 0 and <= 1")
    elif cdf == 0:
        return 0
    cdef float64_t sum2 = 0
    cdef float64_t newval = exp( -1*lam )
    sum2 = newval

    cdef int32_t i
    cdef float64_t sumold
    cdef float64_t lastval

    for i in range(1,maximum+1):
        sumold = sum2
        lastval = newval
        newval = lastval * lam / i
        sum2 = sum2 + newval
        if sumold <= cdf and cdf <= sum2:
            return i

    return maximum

cpdef poisson_cdf_Q_inv ( float64_t cdf, float64_t lam, int32_t maximum=1000 ):
    """inverse poisson distribution.

    cdf : the CDF
    lam : the lambda of poisson distribution

    note: maxmimum return value is 1000
    and lambda must be smaller than 740.
    """
    assert lam < 740
    if cdf < 0 or cdf > 1:
        raise Exception ("CDF must >= 0 and <= 1")
    elif cdf == 0:
        return 0
    cdef float64_t sum2 = 0
    cdef float64_t newval = exp( -1 * lam )
    sum2 = newval

    cdef int32_t i
    cdef float64_t lastval
    cdef float64_t sumold

    for i in range(1,maximum+1):
        sumold = sum2
        lastval = newval
        newval = lastval * lam / i
        sum2 = sum2 + newval
        if sumold <= cdf and cdf <= sum2:
            return i

    return maximum

cpdef poisson_pdf ( uint32_t k, float64_t a ):
    """Poisson PDF.

    PDF(K,A) is the probability that the number of events observed in
    a unit time period will be K, given the expected number of events
    in a unit time as A.
    """
    if a <= 0:
        return 0
    return exp(-a) * pow (a, k, None) / factorial (k)


cdef binomial_coef ( int64_t n, int64_t k ):
    """BINOMIAL_COEF computes the Binomial coefficient C(N,K)

    n,k are integers.
    """
    cdef int64_t mn = min (k, n-k)
    cdef int64_t mx
    cdef float64_t cnk
    cdef int64_t i
    if mn < 0:
        return 0
    elif mn == 0:
        return 1
    else:
        mx = max(k,n-k)
        cnk = <float32_t>(mx+1)
        for i in range(2,mn+1):
            cnk = cnk * (mx+i) / i
    return cnk

cpdef binomial_cdf ( int64_t x, int64_t a, float64_t b, bool lower=True ):
    """ BINOMIAL_CDF compute the binomial CDF.

    CDF(x)(A,B) is the probability of at most X successes in A trials,
    given that the probability of success on a single trial is B.
    """
    if lower:
        return _binomial_cdf_f (x,a,b)
    else:
        return _binomial_cdf_r (x,a,b)

cpdef binomial_sf ( int64_t x, int64_t a, float64_t b, bool lower=True ):
    """ BINOMIAL_SF compute the binomial survival function (1-CDF)

    SF(x)(A,B) is the probability of more than X successes in A trials,
    given that the probability of success on a single trial is B.
    """
    if lower:
        return 1.0 - _binomial_cdf_f (x,a,b)
    else:
        return 1.0 - _binomial_cdf_r (x,a,b)

cpdef pduplication (np.ndarray[np.float64_t] pmf, int32_t N_obs):
    """return the probability of a duplicate fragment given a pmf
    and a number of observed fragments N_obs
    """
    cdef:
        n = pmf.shape[0]
        float32_t p, sf = 0.0
    for p in pmf:
        sf += binomial_sf(2, N_obs, p)
    return sf / <float32_t>n

cdef _binomial_cdf_r ( int64_t x, int64_t a, float64_t b ):
    """ Binomial CDF for upper tail.

    """
    cdef int64_t argmax=<int32_t>(a*b)
    cdef float64_t seedpdf
    cdef float64_t cdf
    cdef float64_t pdf
    cdef int64_t i

    if x < 0:
        return 1
    elif a < x:
        return 0
    elif b == 0:
        return 0
    elif b == 1:
        return 1

    if x<argmax:
        seedpdf=binomial_pdf(argmax,a,b)
        pdf=seedpdf
        cdf = pdf
        for i in range(argmax-1,x,-1):
            pdf/=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf += pdf

        pdf = seedpdf
        i = argmax
        while True:
            pdf*=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf+=pdf
            i+=1
        cdf = min(1,cdf)
        #cdf = float("%.10e" %cdf)
        return cdf
    else:
        pdf=binomial_pdf(x+1,a,b)
        cdf = pdf
        i = x+1
        while True:
            pdf*=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf += pdf
            i+=1
        cdf = min(1,cdf)
        #cdf = float("%.10e" %cdf)
        return cdf


cdef _binomial_cdf_f ( int64_t x, int64_t a, float64_t b ):
    """ Binomial CDF for lower tail.

    """
    cdef int64_t argmax=<int32_t>(a*b)
    cdef float64_t seedpdf
    cdef float64_t cdf
    cdef float64_t pdf
    cdef int64_t i

    if x < 0:
        return 0
    elif a < x:
        return 1
    elif b == 0:
        return 1
    elif b == 1:
        return 0

    if x>argmax:
        seedpdf=binomial_pdf(argmax,a,b)
        pdf=seedpdf
        cdf = pdf
        for i in range(argmax-1,-1,-1):
            pdf/=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf += pdf

        pdf = seedpdf
        for i in range(argmax,x):
            pdf*=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf+=pdf
        cdf=min(1,cdf)
        #cdf = float("%.10e" %cdf)
        return cdf
    else:
        pdf=binomial_pdf(x,a,b)
        cdf = pdf
        for i in range(x-1,-1,-1):
            pdf/=(a-i)*b/(1-b)/(i+1)
            if pdf==0.0: break
            cdf += pdf
        cdf=min(1,cdf)
        #cdf = float("%.10e" %cdf)
        return cdf

cpdef binomial_cdf_inv ( float64_t cdf, int64_t a, float64_t b ):
    """BINOMIAL_CDF_INV inverts the binomial CDF. For lower tail only!

    """
    if cdf < 0 or cdf >1:
        raise Exception("CDF must >= 0 or <= 1")

    cdef float64_t cdf2 = 0
    cdef int64_t x

    for x in range(0,a+1):
        pdf = binomial_pdf (x,a,b)
        cdf2 = cdf2 + pdf
        if cdf < cdf2:
            return x
    return a

cpdef binomial_pdf( int64_t x, int64_t a, float64_t b ):
    """binomial PDF by H. Gene Shin

    """

    if a<1:
        return 0
    elif x<0 or a<x:
        return 0
    elif b==0:
        if x==0:
            return 1
        else:
            return 0
    elif b==1:
        if x==a:
            return 1
        else:
            return 0

    cdef float64_t p
    cdef int64_t mn
    cdef int64_t mx
    cdef float64_t pdf
    cdef int64_t t
    cdef int64_t q

    if x>a-x:
        p=1-b
        mn=a-x
        mx=x
    else:
        p=b
        mn=x
        mx=a-x
    pdf=1
    t = 0
    for q in range(1,mn+1):
        pdf*=(a-q+1)*p/(mn-q+1)
        if pdf < 1e-100:
            while pdf < 1e-3:
                pdf /= 1-p
                t-=1
        if pdf > 1e+100:
            while pdf > 1e+3 and t<mx:
                pdf *= 1-p
                t+=1

    for i in range(mx-t):
        pdf *= 1-p

    #pdf=float("%.10e" % pdf)
    return pdf

# cdef normal_01_cdf ( float64_t x ):
#     """NORMAL_01_CDF evaluates the Normal 01 CDF.
#     """
#     cdef float64_t a1 = 0.398942280444
#     cdef float64_t a2 = 0.399903438504
#     cdef float64_t a3 = 5.75885480458
#     cdef float64_t a4 = 29.8213557808
#     cdef float64_t a5 = 2.62433121679
#     cdef float64_t a6 = 48.6959930692
#     cdef float64_t a7 = 5.92885724438
#     cdef float64_t b0 = 0.398942280385
#     cdef float64_t b1 = 3.8052E-08
#     cdef float64_t b2 = 1.00000615302
#     cdef float64_t b3 = 3.98064794E-04
#     cdef float64_t b4 = 1.98615381364
#     cdef float64_t b5 = 0.151679116635
#     cdef float64_t b6 = 5.29330324926
#     cdef float64_t b7 = 4.8385912808
#     cdef float64_t b8 = 15.1508972451
#     cdef float64_t b9 = 0.742380924027
#     cdef float64_t b10 = 30.789933034
#     cdef float64_t b11 = 3.99019417011
#     cdef float64_t cdf

#     if abs ( x ) <= 1.28:
#         y = 0.5 * x * x
#         q = 0.5 - abs ( x ) * ( a1 - a2 * y / ( y + a3 - a4 / ( y + a5 + a6 / ( y + a7 ) ) ) )
#     elif abs ( x ) <= 12.7:
#         y = 0.5 * x * x

#         q = exp ( - y ) * b0 / ( abs ( x ) - b1 \
#                                 + b2 / ( abs ( x ) + b3 \
#                                 + b4 / ( abs ( x ) - b5 \
#                                 + b6 / ( abs ( x ) + b7 \
#                                 - b8 / ( abs ( x ) + b9 \
#                                 + b10 / ( abs ( x ) + b11 ) ) ) ) ) )
#     else :
#         q = 0.0
#     #
#     #  Take account of negative X.
#     #
#     if x < 0.0:
#         cdf = q
#     else:
#         cdf = 1.0 - q

#     return cdf

# def normal_cdf_inv(p, mu = None, sigma2 = None, lower=True):

#     upper = not lower
#     if p < 0 or p > 1:
#         raise Exception("Illegal argument %f for qnorm(p)." % p)

#     split = 0.42
#     a0 = 2.50662823884
#     a1 = -18.61500062529
#     a2 = 41.39119773534
#     a3 = -25.44106049637
#     b1 = -8.47351093090
#     b2 = 23.08336743743
#     b3 = -21.06224101826
#     b4 = 3.13082909833
#     c0 = -2.78718931138
#     c1 = -2.29796479134
#     c2 = 4.85014127135
#     c3 = 2.32121276858
#     d1 = 3.54388924762
#     d2 = 1.63706781897
#     q = p - 0.5

#     r = 0.0
#     ppnd = 0.0

#     if abs(q) <= split:
#         r = q * q
#         ppnd = q * (((a3 * r + a2) * r + a1) * r + a0) / ((((b4 * r + b3) * r + b2) * r + b1) * r + 1)
#     else:
#         r = p
#         if q > 0:
#             r = 1 - p

#         if r > 0:
#             r = math.sqrt(- math.log(r))
#             ppnd = (((c3 * r + c2) * r + c1) * r + c0) / ((d2 * r + d1) * r + 1)

#             if q < 0:
#                 ppnd = - ppnd
#         else:
#             ppnd = 0

#     if upper:
#         ppnd = - ppnd

#     if mu != None and sigma2 != None:
#         return ppnd * math.sqrt(sigma2) + mu
#     else:
#         return ppnd

# def normal_cdf (z, mu = 0.0, sigma2 = 1.0, lower=True):

#     upper = not lower

#     z = (z - mu) / math.sqrt(sigma2)

#     ltone = 7.0
#     utzero = 18.66
#     con = 1.28
#     a1 = 0.398942280444
#     a2 = 0.399903438504
#     a3 = 5.75885480458
#     a4 = 29.8213557808
#     a5 = 2.62433121679
#     a6 = 48.6959930692
#     a7 = 5.92885724438
#     b1 = 0.398942280385
#     b2 = 3.8052e-8
#     b3 = 1.00000615302
#     b4 = 3.98064794e-4
#     b5 = 1.986153813664
#     b6 = 0.151679116635
#     b7 = 5.29330324926
#     b8 = 4.8385912808
#     b9 = 15.1508972451
#     b10 = 0.742380924027
#     b11 = 30.789933034
#     b12 = 3.99019417011

#     y = 0.0
#     alnorm = 0.0

#     if z < 0:
#         upper = not upper
#         z = - z

#     if z <= ltone or upper and z <= utzero:
#         y = 0.5 * z * z
#         if z > con:
#             alnorm = b1 * math.exp(- y) / (z - b2 + b3 / (z + b4 + b5 / (z - b6 + b7 / (z + b8 - b9 / (z + b10 + b11 / (z + b12))))))
#         else:
#             alnorm = 0.5 - z * (a1 - a2 * y / (y + a3 - a4 / (y + a5 + a6 / (y + a7))))
#     else:
#         alnorm = 0.0
#     if not upper:
#         alnorm = 1.0 - alnorm
#     return alnorm



cimport cython
@cython.boundscheck(False)  # turn off bounds checking
@cython.wraparound(False)  # turn off negative indexing

import warnings
import numpy as np
cimport numpy as np
from itertools import combinations

def match_arg(arg, choices, arg_name):
    assert arg in choices
    return arg

cpdef list match(list a, list b):
    cdef dict b_dict = {x: i for i, x in enumerate(b)}
    return [b_dict.get(x, np.nan) for x in a]

cpdef clipper1sided(np.ndarray[np.float64_t, ndim=2] score_exp, 
                    np.ndarray[np.float64_t, ndim=2] score_back, 
                    int regionLength, double FDR = 0.05, 
                    bint ifuseknockoff = None, 
                    int nknockoff = None, 
                    str contrastScore_method = None, 
                    str importanceScore_method = "diff", 
                    str FDR_control_method = None, 
                    bint ifpowerful = True, 
                    int seed = 12345):
    cdef double shift
    cdef int r1, r2
    cdef bint re

    score_exp = np.atleast_2d(score_exp)
    score_back = np.atleast_2d(score_back)

    if np.any(score_exp < 0) or np.any(score_back < 0):
        shift = min(np.min(score_exp[~np.isnan(score_exp)]), np.min(score_back[~np.isnan(score_back)]))
        score_exp -= shift
        score_back -= shift
  
    r1 = np.shape(score_exp)[1]
    r2 = np.shape(score_back)[1]

    if np.shape(score_exp)[0] != np.shape(score_back)[0]:
        sys.exit("score_exp and score_back must have the same number of rows (features)")
    
    if ifuseknockoff is None:
        ifuseknockoff = (r1 != r2)

    if not ifuseknockoff:
        if r1 != r2:
            warnings.warn("Caution: no knockoffs are constructed when the numbers of replicates are different; FDR control is not guaranteed")
        if FDR_control_method == "GZ":
            warnings.warn("FDR_control_method cannot be GZ when no knockoffs are constructed. Switching to BC.")
        FDR_control_method = "BC"
        warnings.simplefilter("ignore")
        re = clipper_1sided_woknockoff(score_exp=score_exp, 
                                    score_back=score_back, 
                                    r1=r1, r2=r2, 
                                    FDR=FDR,
                                    importanceScore_method=importanceScore_method,
                                    FDR_control_method=FDR_control_method,
                                    regionLength=regionLength)
    
    if ifuseknockoff:
        if r1 == 1 and r2 == 1:
            sys.exit("Cannot generate knockoffs when both score_exp and score_back have one column. Please rerun clipper1sided by setting ifuseknockoff = F")
        nknockoff_max = len(list(combinations(range(r1+r2), r1)))/2 - 1 if r1 == r2 else len(list(combinations(range(r1+r2), r1))) - 1
        if nknockoff is not None:
            if nknockoff > nknockoff_max or (not isinstance(nknockoff, int)) or nknockoff < 1:
                warnings.warn("nknockoff must be a positive integer and must not exceed the maximum number of knockoff; using the exhaustive knockoffs instead.")
            nknockoff = min(nknockoff, nknockoff_max)
        else:
            warnings.warn(f"nknockoff is not supplied; generate the maximum number of knockoffs: {nknockoff_max}")
            nknockoff = nknockoff_max if contrastScore_method == "diff" else 1
        warnings.simplefilter("ignore")
        re = clipper_1sided_wknockoff(score_exp=score_exp,
                                    score_back=score_back,
                                    r1=r1, r2=r2,
                                    FDR=FDR,
                                    importanceScore_method=importanceScore_method,
                                    contrastScore_method= contrastScore_method,
                                    nknockoff=nknockoff,
                                    nknockoff_max=nknockoff_max,
                                    seed=seed)
    
    if ifpowerful and FDR_control_method != "BH":
        FDR_nodisc = np.array(list(map(lambda re_i: len(re_i["discovery"]) == 0, re["results"])))
        if np.any(FDR_nodisc):
            warnings.warn(f"At FDR = {', '.join([str(x) for x in FDR[FDR_nodisc]])}, no discovery has been found using FDR control method {FDR_control_method}; switching to BH...")
            re_clipperbh = clipper_BH(contrastScore=re["contrastScore"], nknockoff=nknockoff, FDR=FDR[FDR_nodisc])
            re["results"][FDR_nodisc] = re_clipperbh
        
    return re

def clipper2sided(score_exp, 
                  score_back, 
                  FDR = 0.05, 
                  ifuseknockoff = None,
                  nknockoff = None,
                  contrastScore_method = "max",
                  importanceScore_method = "diff",
                  FDR_control_method = "GZ",
                  ifpowerful = True,
                  seed = 12345):

cpdef dict clipper2sided(np.ndarray[np.float64_t, ndim=2] score_exp, 
                         np.ndarray[np.float64_t, ndim=2] score_back, 
                         double FDR = 0.05,
                         object ifuseknockoff = None, 
                         int nknockoff = None, 
                         str contrastScore_method = "max", 
                         str importanceScore_method = "diff", 
                         str FDR_control_method = "GZ", 
                         bint ifpowerful = True, 
                         int seed = 12345):
    
    cdef int r1 = score_exp.shape[1]
    cdef int r2 = score_back.shape[1]
    cdef int nknockoff_max
    cdef object knockoffidx, kappatau_ls, re
    cdef np.ndarray[np.int_t, ndim=1] FDR_nodisc
    cdef dict re_clipperbh
    
    score_exp = np.atleast_2d(score_exp)
    score_back = np.atleast_2d(score_back)

    r1 = np.shape(score_exp)[1]
    r2 = np.shape(score_back)[1]

    if r1 == 1 and r2 == 1:
        sys.exit("clipper is not yet able to perform two sided identification when either condition has one replicate")
  
    nknockoff_max = (len(list(combinations(range(r1+r2), r1)))/2 - 1) if r1 == r2 else (len(list(combinations(range(r1+r2), r1))) - 1)
    nknockoff_max = int(min(nknockoff_max, 200))
    if nknockoff is not None:
        if nknockoff > nknockoff_max or (not isinstance(nknockoff, int)) or nknockoff < 1:
            warnings.warn("nknockoff must be a positive integer and must not exceed the maximum number of knockoff; using the maximal number of knockoffs instead.")
        nknockoff = min(nknockoff, nknockoff_max)
    else:
        nknockoff = min(nknockoff_max, 50) if contrastScore_method == "diff" else 1

    knockoffidx = generate_knockoffidx(r1=r1, r2=r2, 
                                        nknockoff=nknockoff, 
                                        nknockoff_max=nknockoff_max,
                                        seed=seed)

    kappatau_ls = compute_taukappa(score_exp=score_exp,
                                    score_back=score_back,
                                    r1=r1, r2=r2,
                                    if2sided=True,
                                    knockoffidx=knockoffidx,
                                    importanceScore_method=importanceScore_method,
                                    contrastScore_method=contrastScore_method)
    re = clipper_GZ(tau=kappatau_ls["tau"], kappa=kappatau_ls["kappa"],
                    nknockoff=nknockoff, FDR=FDR)
    re = {"knockoffidx": knockoffidx,
            "importanceScore_method": importanceScore_method,
            "importanceScore": kappatau_ls["importanceScore"],
            "contrastScore_method": contrastScore_method,
            "contrastScore": (2*kappatau_ls["kappa"] - 1) * abs(kappatau_ls["tau"]),
            "results": re}

    if ifpowerful and FDR_control_method != "BH":
        FDR_nodisc = np.array(list(map(lambda re_i: len(re_i["discovery"]) == 0, re["results"])))

        if np.any(FDR_nodisc & (contrastScore_method == "max")):
            warnings.warn(f"At FDR = {', '.join([str(x) for x in FDR[FDR_nodisc]])}, no discovery has been found using FDR control method {FDR_control_method}; switching to BH...")
            re_clipperbh = clipper_BH(contrastScore=re["contrastScore"], nknockoff=nknockoff, FDR=FDR[FDR_nodisc])
            re["results"][FDR_nodisc] = re_clipperbh
    
    return re

def clipper_1sided_woknockoff(score_exp,
                              score_back,
                              r1, r2,
                              importanceScore_method,
                              FDR_control_method,
                              regionLength,
                              FDR = 0.05,
                              aggregation_method = "mean"):
    if r1 > 1:
        score_exp = aggregate_clipper(score=score_exp, aggregation_method=aggregation_method)
    if r2 > 1:
        score_back = aggregate_clipper(score=score_back, aggregation_method=aggregation_method)
    contrastscore = compute_importanceScore_wsinglerep(score_exp=score_exp,
                                                     score_back=score_back,
                                                     importanceScore_method=importanceScore_method)
    if FDR_control_method == "BC":
        warnings.simplefilter("ignore")
        re = clipper_BC(contrastScore=contrastscore, FDR=FDR, regionLength=regionLength)
  
    if FDR_control_method == "BH":
        re = clipper_BH(contrastscore=contrastscore, FDR=FDR)
  
    re = {"importanceScore": contrastscore,
            "importanceScore_method": importanceScore_method,
            "contrastScore": contrastscore,
            "contrastScore_method": importanceScore_method,
            "results": re}
    return re

def clipper_1sided_wknockoff(score_exp,
                             score_back,
                             r1, r2,
                             importanceScore_method,
                             contrastScore_method,
                             nknockoff,
                             nknockoff_max,
                             seed,
                             FDR = 0.05):
    knockoffidx = generate_knockoffidx(r1=r1, r2=r2, 
                                        nknockoff=nknockoff, 
                                        nknockoff_max=nknockoff_max,
                                        seed=seed)
    kappatau_ls = compute_taukappa(score_exp=score_exp,
                                    score_back=score_back,
                                    r1=r1, r2=r2,
                                    if2sided=True,
                                    knockoffidx=knockoffidx,
                                    importanceScore_method=importanceScore_method,
                                    contrastScore_method=contrastScore_method)
    re = clipper_GZ(tau=kappatau_ls["tau"], kappa=kappatau_ls["kappa"], nknockoff=nknockoff, FDR=FDR)
    re = {"knockoffidx": knockoffidx,
            "importanceScore_method": importanceScore_method,
            "importanceScore": kappatau_ls["importanceScore"],
            "contrastScore_method": contrastScore_method,
            "contrastScore": (2*kappatau_ls["kappa"]-1)*abs(kappatau_ls["tau"]),
            "results": re}
    return re


def aggregate_clipper(score, aggregation_method):
    if aggregation_method == "mean":
        score_single = np.array(list(map(np.nanmean, score)))
    if aggregation_method == "median":
        score_single = np.array(list(map(np.nanmedian, score)))
    return score_single

def compute_importanceScore_wsinglerep(score_exp, score_back, importanceScore_method):
    if importanceScore_method == "diff":
        contrastScore = score_exp - score_back
    if importanceScore_method == "max":
        contrastScore = np.maximum(score_exp, score_back)
    return np.array(contrastScore).flatten()

# def clipper_BC(contrastScore, FDR):
#     contrastScore[np.isnan(contrastScore)] = 0
#     c_abs = np.sort(np.unique(np.abs(contrastScore[contrastScore != 0])))
#     i = 0
#     emp_fdp = np.repeat(np.nan, len(c_abs))
#     emp_fdp[0] = 1
#
#     print('contrastScore:')
#     print(contrastScore)
#     # b[np.where(a == value)]
#
#     while (i < len(c_abs)):
#         t = c_abs[i]
#         emp_fdp[i] = min((1 + np.sum(contrastScore <= -t))/np.sum(contrastScore >= t), 1)
#         if i >= 1:
#             emp_fdp[i] = min(emp_fdp[i], emp_fdp[i-1])
#         i += 1
#
#     c_abs = c_abs[~np.isnan(emp_fdp)]
#     emp_fdp = emp_fdp[~np.isnan(emp_fdp)]
#
#     q_idx = match(contrastScore, c_abs)
#     q = np.array(list(map(lambda k: emp_fdp[k] if k == k else 1, q_idx)))
#
#     def temp(FDR_i):
#         try:
#             thre = c_abs[np.min(np.where(emp_fdp <= FDR_i))]
#         except:
#             thre = np.nan
#         return {"FDR": FDR_i,
#                     "FDR_control": "BC",
#                     "thre": thre,
#                     "q": q,
#                     "discovery": np.where(contrastScore >= thre)[0]
#                     }
#
#     re = np.array(list(map(temp, FDR)))
#     return re

#################################### Add length information (by Saidi) #################################################
def clipper_BC(contrastScore, FDR, regionLength):
    contrastScore[np.isnan(contrastScore)] = 0
    c_abs = np.sort(np.unique(np.abs(contrastScore[contrastScore != 0])))
    i = 0
    emp_fdp = np.repeat(np.nan, len(c_abs))
    emp_fdp[0] = 1

    # print('contrastScore:')
    # print(contrastScore)
    # b[np.where(a == value)]
    # print('regionLength')
    # print(regionLength)
    while (i < len(c_abs)):
        t = c_abs[i]

        # emp_fdp[i] = min((1 + np.sum(contrastScore <= -t)) / np.sum(contrastScore >= t), 1)
        ### Add the length information by Saidi##################
        emp_fdp[i] = min((1 + sum(regionLength[np.where(contrastScore <= -t)])) / sum(regionLength[np.where(contrastScore >= t)]), 1)
        # print(emp_fdp[i])
        if i >= 1:
            emp_fdp[i] = min(emp_fdp[i], emp_fdp[i - 1])
        i += 1
        np.array(emp_fdp))

    c_abs = c_abs[~np.isnan(emp_fdp)]
    emp_fdp = emp_fdp[~np.isnan(emp_fdp)]

    q_idx = match(contrastScore, c_abs)
    q = np.array(list(map(lambda k: emp_fdp[k] if k == k else 1, q_idx)))

    def temp(FDR_i):
        try:
            thre = c_abs[np.min(np.where(emp_fdp <= FDR_i))]
        except:
            thre = np.nan
        return {"FDR": FDR_i,
                "FDR_control": "BC",
                "thre": thre,
                "q": q,
                "discovery": np.where(contrastScore >= thre)[0]
                }

    re = np.array(list(map(temp, FDR)))
    return re


def clipper_BH(contrastScore, FDR, nknockoff = None):
    if isinstance(contrastScore, dict):
            n = len(np.atleast_1d(contrastScore[1]))
            kappa = contrastScore["kappa"]
            tau = contrastScore["tau"]
            idx_na = np.isnan(tau) | np.isnan(kappa)
            tau = tau[~idx_na]
            kappa = kappa[~idx_na]
            def temp(i):
                t1 = np.array((~kappa) & tau >= tau[i])
                t1 = t1[~np.isnan(t1)]
                t2 = np.array(~kappa)
                t2 = t2[~np.isnan(t2)]
                return sum(t1) / sum(t2) * nknockoff / (nknockoff + 1)
            pval = np.array(list(map(temp, range(n))))
    else:
            contrastScore = np.atleast_1d(contrastScore)
            n = len(contrastScore)
            idx_na = np.isnan(contrastScore)
            contrastScore_nomiss = contrastScore[~idx_na]
            cs_negative = contrastScore_nomiss[contrastScore_nomiss < 0]
            cs_null = np.concatenate((cs_negative, -cs_negative))
            pval = np.array(list(map(lambda x: np.mean(x <= cs_null), contrastScore_nomiss)))

    # qvalue = np.array(stats.p_adjust(FloatVector(pval), method = 'BH'))
    qvalue = pval
    re = np.array(list(map(lambda FDR_i: 
                            {"FDR": FDR,
                                "FDR_control": "BH",
                                "discovery": np.arange(n)[~idx_na][np.where(qvalue <= FDR_i)[0]],
                                "q": qvalue},FDR)))
    return re


def generate_knockoffidx(r1, r2, nknockoff, nknockoff_max, seed):

    np.random.seed(seed)
    if nknockoff_max == 200:
            knockoffidx = [np.nan]*nknockoff
            i_knockoff = 0
            while (i_knockoff < nknockoff):
                temp = np.random.choice(r1+r2, r1, replace=False)
                if np.any(~np.isin(temp, np.arange(r1))) and np.any(np.isin(temp, np.arange(r1))):
                        knockoffidx[i_knockoff] = temp
                        i_knockoff += 1
                else:
                        continue
    else:
        combination_all = np.array([list(x) for x in combinations(range(r1 + r2), r1)], dtype=object)

        if r1 == r2:
            combination_all = combination_all[:len(combination_all)//2][1:]
        else:
            combination_all = combination_all[1:]

        knockoffidx = combination_all[np.random.choice(nknockoff_max, size=nknockoff, replace=False)].tolist()

    return knockoffidx

def compute_taukappa(score_exp, score_back, r1, r2, if2sided,
                     knockoffidx, importanceScore_method, contrastScore_method):

    perm_idx = knockoffidx.copy()
    perm_idx.insert(0, list(range(r1)))

    score_tot = np.concatenate((score_exp, score_back), axis=1)

    def imp_ls_apply(x):
        se = score_tot[:, x]
        sb = score_tot[:, np.setdiff1d(np.arange(r1+r2), x)]
        se = aggregate_clipper(se, aggregation_method='mean')
        sb = aggregate_clipper(sb, aggregation_method='mean')
        imp = compute_importanceScore_wsinglerep(se, sb, importanceScore_method=importanceScore_method)
        if if2sided: imp = np.abs(imp)
        return imp
    imp_ls = np.array(list(map(imp_ls_apply, perm_idx)), dtype=object).T
    def kappatau_ls_apply(x):
        kappa = ~np.any(x[1:] == np.max(x))
        if len(np.atleast_1d(kappa)) == 0:
            kappa = np.nan
        x_sorted = np.sort(x)[::-1]
        if contrastScore_method == "diff":
            tau = x_sorted[0] - x_sorted[1]
        if contrastScore_method == "max":
            tau = x_sorted[0]
        return {"kappa": kappa, "tau": tau}
    kappatau_ls = list(map(kappatau_ls_apply, imp_ls))

    re = {"importanceScore": imp_ls,
            "kappa": np.array([kappatau["kappa"] for kappatau in kappatau_ls]),
            "tau": np.array([kappatau["tau"] for kappatau in kappatau_ls])}
    return re

def clipper_GZ(tau, kappa, nknockoff, FDR):

    contrastScore = (2 * kappa - 1) * np.abs(tau)
    contrastScore[np.isnan(contrastScore)] = 0
    c_abs = np.abs(contrastScore[contrastScore != 0])
    c_abs = np.sort(np.unique(c_abs))

    i = 0
    emp_fdp = np.repeat(np.nan, len(c_abs))
    emp_fdp[0] = 1
    while (i < len(c_abs)):
            t = c_abs[i]
            emp_fdp[i] = min((1/nknockoff + 1/nknockoff * np.sum(contrastScore <= -t))/np.sum(contrastScore >= t), 1)
            if i >= 1: emp_fdp[i] = min(emp_fdp[i], emp_fdp[i-1])
            i += 1
            
    c_abs = c_abs[~np.isnan(emp_fdp)]
    emp_fdp = emp_fdp[~np.isnan(emp_fdp)]
    q_idx = match(contrastScore, c_abs)
    q = np.array(list(map(lambda k: emp_fdp[k] if k == k else 1, q_idx)))

    def temp(FDR_i):
        try:
            thre = c_abs[np.min(np.where(emp_fdp <= FDR_i))]
        except:
            thre = np.nan
        return {"FDR": FDR_i,
                "FDR_control": "BC",
                "thre": thre,
                "q": q,
                "discovery": np.where(contrastScore >= thre)[0]
                }
    re = np.array(list(map(temp, FDR)))
    return re


def clipper(score_exp, score_back, analysis, regionLength, FDR = 0.05,
            procedure = None, contrast_score = None,
            num_permutation = None, seed = 12345):

    analysis = match_arg(analysis, ["differential", "enrichment"], "analysis")

    score_exp = np.atleast_2d(score_exp)
    score_back = np.atleast_2d(score_back)

    FDR = np.atleast_1d(FDR)
    if (procedure is None):
        procedure = "GZ" if analysis == "differential" else "BC"
    else:
        procedure = match_arg(procedure, ["BC", "aBH", "GZ"], "procedure")

    if (analysis == "differential"):
        contrast_score = match_arg(contrast_score, ["diff", "max"], "contrast_score") if contrast_score is not None else "max"

        re = clipper2sided(score_exp=score_exp, score_back=score_back, FDR=FDR,
                        nknockoff=num_permutation, 
                        contrastScore_method= contrast_score, importanceScore_method="diff",
                        FDR_control_method=procedure, ifpowerful=False, seed=seed)
        
        FDR_nodisc = np.array(list(map(lambda re_i: len(re_i["discovery"]) == 0, re["results"])))
        if np.any(np.array(FDR_nodisc) & (contrast_score == "max")):
            warnings.warn(f"At FDR = {', '.join([str(x) for x in FDR[FDR_nodisc]])}, no discovery has been found using max contrast score. To make more discoveries, switch to diff contrast score or increase the FDR threshold. ")

    elif (analysis == "enrichment"):

        if (np.shape(score_exp)[1] != np.shape(score_back)[1]):
            procedure = "GZ"
        if (contrast_score is None):
            contrast_score = "diff" if procedure == "BC" else "max"
        else:
            contrast_score = match_arg(contrast_score, ["diff", "max"], "contrast_score")
        if procedure == "BC":
            re = clipper1sided(score_exp=score_exp, score_back=score_back, FDR=FDR,
                            nknockoff=num_permutation,
                            importanceScore_method=contrast_score,
                            FDR_control_method=procedure, ifpowerful=False, seed=seed,regionLength=regionLength)
        if procedure == "GZ":
            re = clipper1sided(score_exp=score_exp, score_back=score_back, FDR=FDR,
                            nknockoff=num_permutation,
                            contrastScore_method=contrast_score,
                            FDR_control_method=procedure, ifpowerful=False, seed=seed,regionLength=regionLength)

        FDR_nodisc = np.array(list(map(lambda re_i: len(re_i["discovery"]) == 0, re["results"])))
        if np.any(np.array(FDR_nodisc) & (procedure != 'aBH')):
            warnings.warn(f"At FDR = {', '.join([str(x) for x in FDR[FDR_nodisc]])}, no discovery has been found using max contrast score. To make more discoveries, switch to diff contrast score or increase the FDR threshold. ")

    contrast_score_value = re["contrastScore"]
    thre = []
    discoveries = []
    results = re["results"]
    for result in results:
        try:
            thre.append(result["thre"])
            discoveries.append(result["discovery"])
        except:
            continue
    thre = np.array(thre)
    discoveries = np.array(discoveries)
    q = results[0]["q"]
    re = {"contrast.score": contrast_score, 
            "contrast.score.value": contrast_score_value,
            "FDR": FDR,
            "contrast.score.thre": thre,
            "discoveries": discoveries,
            "q": q}

    print(f"thre:{thre}")

    return re


# cython: language_level=3
# cython: profile=True
# Time-stamp: <2023-11-02 11:40:47 Tao Liu>

"""Module for BedGraph data class.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
#from array import array
from cpython cimport array
from array import array as pyarray
from math import prod
# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.Signal.ScoreTrack import ScoreTrackII
from MACS3.IO.PeakIO import PeakIO, BroadPeakIO, RegionIO
from MACS3.Signal.Prob import chisq_logp_e

# ------------------------------------
# Other modules
# ------------------------------------

from cpython cimport bool
import numpy as np
cimport numpy as np
from numpy cimport uint8_t, uint16_t, uint32_t, uint64_t, int8_t, int16_t, int32_t, int64_t, float32_t, float64_t

# ------------------------------------
# C lib
# ------------------------------------

from libc.math cimport sqrt, log, log1p, exp, log10

# ------------------------------------
# constants
# ------------------------------------
__version__ = "BedGraph $Revision$"
__author__ = "Tao Liu <vladimir.liu@gmail.com>"
__doc__ = "bedGraphTrackI class"

# ------------------------------------
# Misc functions
# ------------------------------------
LOG10_E = 0.43429448190325176

cdef inline mean_func( x ):
    return sum( x )/len( x )

cdef inline fisher_func( x ):
    # combine -log10pvalues
    return chisq_logp_e( 2*sum (x )/LOG10_E, 2*len( x ), log10=True )

cdef inline subtract_func( x ):
    # subtraction of two items list
    return x[1] - x[0]

cdef inline divide_func( x ):
    # division of two items list
    return x[1] / x[2]

cdef inline product_func( x ):
    # production of a list of values
    # only python 3.8 or above
    return prod( x )
    
# ------------------------------------
# Classes
# ------------------------------------
cdef class bedGraphTrackI:
    """Class for bedGraph type data.

    In bedGraph, data are represented as continuous non-overlapping
    regions in the whole genome. I keep this assumption in all the
    functions. If data has overlaps, some functions will definitely
    give incorrect results.

    1. Continuous: the next region should be after the previous one
    unless they are on different chromosomes;

    2. Non-overlapping: the next region should never have overlaps
    with preceding region.

    The way to memorize bedGraph data is to remember the transition
    points together with values of their preceding regions. The last
    data point may exceed chromosome end, unless a chromosome
    dictionary is given. Remember the coordinations in bedGraph and
    this class is 0-indexed and right-open.

    """
    cdef:
        dict __data
        public float32_t maxvalue
        public float32_t minvalue
        public float32_t baseline_value

    def __init__ (self, float32_t baseline_value=0 ):
        """
        baseline_value is the value to fill in the regions not defined
        in bedGraph. For example, if the bedGraph is like:

        chr1  100 200  1
        chr1  250 350  2

        Then the region chr1:200..250 should be filled with baseline_value.

        """
        self.__data = {}
        self.maxvalue = -10000000 # initial maximum value is tiny since I want safe_add_loc to update it
        self.minvalue = 10000000  # initial minimum value is large since I want safe_add_loc to update it
        self.baseline_value = baseline_value

    cpdef add_loc ( self, bytes chromosome, int32_t startpos, int32_t endpos, float32_t value):
        """Add a chr-start-end-value block into __data dictionary.

        Difference between safe_add_loc: no check, but faster. Save
        time while being called purely within MACS, so that regions
        are continuous without gaps. Note: I forgot that I removed
        safe_add_loc :P

        """
        cdef float32_t pre_v
        # basic assumption, end pos should > start pos

        if endpos <= 0:
            return
        if startpos < 0:
            startpos = 0

        if chromosome not in self.__data:
            self.__data[chromosome] = [ pyarray('i',[]), pyarray('f',[]) ]
            c = self.__data[chromosome]
            if startpos:
                # start pos is not 0, then add two blocks, the first
                # with "baseline_value"; the second with "value"
                c[0].append(startpos)
                c[1].append(self.baseline_value)
            c[0].append(endpos)
            c[1].append(value)
        else:
            c = self.__data[chromosome]
            # get the preceding region
            pre_v   = c[1][-1]

            # if this region is next to the previous one.
            if pre_v == value:
                # if value is the same, simply extend it.
                c[0][-1] = endpos
            else:
                # otherwise, add a new region
                c[0].append(endpos)
                c[1].append(value)

        if value > self.maxvalue:
            self.maxvalue = value
        if value < self.minvalue:
            self.minvalue = value

    cpdef add_loc_wo_merge ( self, bytes chromosome, int32_t startpos, int32_t endpos, float32_t value):
        """Add a chr-start-end-value block into __data dictionary.

        Difference between safe_add_loc: no check, but faster. Save
        time while being called purely within MACS, so that regions
        are continuous without gaps. Note: I forgot that I removed
        safe_add_loc :P

        This one won't merge nearby ranges with the same value
        """
        if endpos <= 0:
            return
        if startpos < 0:
            startpos = 0

        if value < self.baseline_value:
            value = self.baseline_value
            
        if chromosome not in self.__data:
            self.__data[chromosome] = [ pyarray('i',[]), pyarray('f',[]) ]
            c = self.__data[chromosome]
            if startpos:
                # start pos is not 0, then add two blocks, the first
                # with "baseline_value"; the second with "value"
                c[0].append(startpos)
                c[1].append(self.baseline_value)
        c = self.__data[chromosome]
        c[0].append(endpos)
        c[1].append(value)
        if value > self.maxvalue:
            self.maxvalue = value
        if value < self.minvalue:
            self.minvalue = value

    cpdef add_chrom_data( self, bytes chromosome, object p, object v ):
        """Add a pv data to a chromosome. Replace the previous data.

        p: a pyarray object 'i' for positions
        v: a pyarray object 'f' for values

        Note: no checks for error, use with caution
        """
        cdef:
            float32_t maxv, minv

        self.__data[ chromosome ] = [ p, v ]
        maxv = max( v )
        minv = min( v )
        if maxv > self.maxvalue:
            self.maxvalue = maxv
        if minv < self.minvalue:
            self.minvalue = minv
        return

    cpdef add_chrom_data_hmmr_PV( self, bytes chromosome, object pv ):
        """Add a pv data to a chromosome. Replace the previous data.

        This is a kinda silly function to waste time and convert a PV
        array (2-d named numpy array) into two python arrays for this
        BedGraph class. May have better function later.

        Note: no checks for error, use with caution
        """
        cdef:
            float32_t maxv, minv
            int32_t i

        self.__data[ chromosome ] = [ pyarray('i', pv['p']), pyarray('f',pv['v']) ]
        minv = pv['v'].min()
        maxv = pv['p'].max()
        if maxv > self.maxvalue:
            self.maxvalue = maxv
        if minv < self.minvalue:
            self.minvalue = minv
        return
    
    cpdef bool destroy ( self ):
        """ destroy content, free memory.
        """
        cdef:
            set chrs
            bytes chrom

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            if chrom in self.__data:
                self.__data[chrom] = [None, None]
                self.__data.pop(chrom)
        return True

    cpdef list get_data_by_chr (self, bytes chromosome):
        """Return array of counts by chromosome.

        The return value is a tuple:
        ([end pos],[value])
        """
        if chromosome in self.__data:
            return self.__data[chromosome]
        else:
            return []

    cpdef set get_chr_names (self):
        """Return all the chromosome names stored.

        """
        return set(sorted(self.__data.keys()))

    cpdef void write_bedGraph (self, fhd, str name, str description, bool trackline=True):
        """Write all data to fhd in Wiggle Format.

        fhd: a filehandler to save bedGraph.
        name/description: the name and description in track line.

        shift will be used to shift the coordinates. default: 0
        """
        cdef:
            int32_t pre, pos, i
            float32_t value
            bytes chrom
            set chrs

        if trackline:
            trackcontents = (name.replace("\"", "\\\""), description.replace("\"", "\\\""))
            fhd.write("track type=bedGraph name=\"%s\" description=\"%s\" visibility=2 alwaysZero=on\n" % trackcontents)
        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            (p,v) = self.__data[chrom]
            pnext = iter(p).__next__
            vnext = iter(v).__next__
            pre = 0

            for i in range(len(p)):
                pos = pnext()
                value = vnext()
                #if value != self.baseline_value:
                # never write baseline_value
                fhd.write("%s\t%d\t%d\t%.5f\n" % (chrom.decode(),pre,pos,value))
                pre = pos
        return

    cpdef void reset_baseline (self, float32_t baseline_value):
        """Reset baseline value to baseline_value.

        So any region between self.baseline_value and baseline_value
        will be set to baseline_value.

        """
        self.baseline_value = baseline_value
        self.filter_score(cutoff=baseline_value)
        self.merge_regions()
        return

    cdef merge_regions (self):
        """Merge nearby regions with the same value.

        """
        cdef:
            int32_t new_pre_pos, pos, i
            float32_t new_pre_value, value
            bytes chrom
            set chrs

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            (p,v) = self.__data[chrom]
            pnext = iter(p).__next__
            vnext = iter(v).__next__

            # new arrays
            new_pos = pyarray('L',[pnext(),])
            new_value = pyarray('f',[vnext(),])

            newpa = new_pos.append
            newva = new_value.append

            new_pre_pos = new_pos[0]
            new_pre_value = new_value[0]

            for i in range(1,len(p)):
                pos = pnext()
                value = vnext()
                if value == new_pre_value:
                    new_pos[-1] = pos
                else:
                    # add new region
                    newpa(pos)
                    newva(value)
                    new_pre_pos = pos
                    new_pre_value = value
            self.__data[chrom] = [new_pos,new_value]
        return True

    cpdef bool filter_score (self, float32_t cutoff=0):
        """Filter using a score cutoff. Any region lower than score
        cutoff will be set to self.baseline_value.

        Self will be modified.
        """
        cdef:
            int32_t new_pre_pos, pos, i
            float32_t new_pre_value, value
            bytes chrom
            set chrs

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            (p,v) = self.__data[chrom]
            pnext = iter(p).__next__
            vnext = iter(v).__next__

            # new arrays
            new_pos = pyarray('L',[])
            new_value = pyarray('f',[])
            new_pre_pos = 0
            new_pre_value = 0

            for i in range(len(p)):
                pos = pnext()
                value = vnext()

                if value < cutoff:
                    # this region will be set to baseline_value
                    if new_pre_value == self.baseline_value:
                        # if preceding region is at baseline, extend it
                        new_pos[-1] = pos
                    else:
                        # else add a new baseline region
                        new_pos.append(pos)
                        new_value.append(self.baseline_value)
                else:
                    # put it into new arrays
                    new_pos.append(pos)
                    new_value.append(value)
                new_pre_pos = new_pos[-1]
                new_pre_value = new_value[-1]
            self.__data[chrom]=[new_pos,new_value]
        return True

    cpdef tuple summary (self):
        """Calculate the sum, total_length, max, min, mean, and std. 

        Return a tuple for (sum, total_length, max, min, mean, std).
        """
        cdef:
            int64_tn_v
            float32_t sum_v, max_v, min_v, mean_v, variance, tmp, std_v
            int32_t pre_p, l, i

        pre_p = 0
        n_v = 0
        sum_v = 0
        max_v = -100000
        min_v = 100000
        for (p,v) in self.__data.values():
            # for each chromosome
            pre_p = 0
            for i in range(len(p)):
                # for each region
                l = p[i]-pre_p
                sum_v += v[i]*l
                n_v += l
                pre_p = p[i]
            max_v = max(max(v),max_v)
            min_v = min(min(v),min_v)
        mean_v = sum_v/n_v
        variance = 0.0
        for (p,v) in self.__data.values():
            for i in range(len(p)):
                # for each region
                tmp = v[i]-mean_v
                l = p[i]-pre_p
                variance += tmp*tmp*l
                pre_p = p[i]

        variance /= float(n_v-1)
        std_v = sqrt(variance)
        return (sum_v, n_v, max_v, min_v, mean_v, std_v)

    cpdef object call_peaks (self, float32_t cutoff=1,
                             int32_t min_length=200, int32_t max_gap=50,
                             bool call_summits=False):
        """This function try to find regions within which, scores
        are continuously higher than a given cutoff.

        This function is NOT using sliding-windows. Instead, any
        regions in bedGraph above certain cutoff will be detected,
        then merged if the gap between nearby two regions are below
        max_gap. After this, peak is reported if its length is above
        min_length.

        cutoff:  cutoff of value, default 1.
        min_length :  minimum peak length, default 200.
        gap   :  maximum gap to merge nearby peaks, default 50.

        Removed option:

        up_limit: the highest acceptable value. Default 10^{310}
          * so only allow peak with value >=cutoff and <=up_limit

        This does not work. The region above upper limit may still be
        included as `gap` .

        """
        cdef:
            int32_t peak_length, x, pre_p, p, i, summit, tstart, tend
            float32_t v, summit_value, tvalue
            bytes chrom
            set chrs
            object peaks

        chrs = self.get_chr_names()
        peaks = PeakIO()                      # dictionary to save peaks
        for chrom in sorted(chrs):
            peak_content = None
            peak_length = 0
            (ps,vs) = self.get_data_by_chr(chrom) # arrays for position and values
            psn = iter(ps).__next__         # assign the next function to a viable to speed up
            vsn = iter(vs).__next__
            x = 0
            pre_p = 0                   # remember previous position
            while True:
                # find the first region above cutoff
                try:                    # try to read the first data range for this chrom
                    p = psn()
                    v = vsn()
                except:
                    break
                x += 1                  # index for the next point
                if v >= cutoff:
                    peak_content = [(pre_p,p,v),]
                    pre_p = p
                    break               # found the first range above cutoff
                else:
                    pre_p = p

            for i in range(x,len(ps)):
                # continue scan the rest regions
                p = psn()
                v = vsn()
                if v < cutoff: # not be detected as 'peak'
                    pre_p = p
                    continue
                # for points above cutoff
                # if the gap is allowed
                if pre_p - peak_content[-1][1] <= max_gap:
                    peak_content.append((pre_p,p,v))
                else:
                    # when the gap is not allowed, close this peak
                    self.__close_peak(peak_content, peaks, min_length, chrom) #, smoothlen=max_gap / 2 )
                    # start a new peak
                    peak_content = [(pre_p,p,v),]
                pre_p = p

            # save the last peak
            if not peak_content:
                continue
            self.__close_peak(peak_content, peaks, min_length, chrom) #, smoothlen=max_gap / 2 )
        return peaks

    cdef bool __close_peak( self, list peak_content, object peaks, int32_t min_length, bytes chrom ):
        cdef:
            list tsummit        # list for temporary summits
            int32_t peak_length, summit, tstart, tend
            float32_t summit_value, tvalue
            
        peak_length = peak_content[-1][1]-peak_content[0][0]
        if peak_length >= min_length: # if the peak is too small, reject it
            tsummit = []
            summit = 0
            summit_value = 0
            for (tstart,tend,tvalue) in peak_content:
                if not summit_value or summit_value < tvalue:
                    tsummit = [<int32_t>((tend+tstart)/2),]
                    summit_value = tvalue
                elif summit_value == tvalue:
                    tsummit.append( <int32_t>((tend+tstart)/2) )
            summit = tsummit[<int32_t>((len(tsummit)+1)/2)-1 ]
            peaks.add( chrom,
                       peak_content[0][0],
                       peak_content[-1][1],
                       summit      = summit,
                       peak_score  = summit_value,
                       pileup      = 0,
                       pscore      = 0,
                       fold_change = 0,
                       qscore      = 0
                       )
            return True

    cpdef object call_broadpeaks (self, float32_t lvl1_cutoff=500, float32_t lvl2_cutoff=100,
                                  int32_t min_length=200, int32_t lvl1_max_gap=50, int32_t lvl2_max_gap=400):
        """This function try to find enriched regions within which,
        scores are continuously higher than a given cutoff for level
        1, and link them using the gap above level 2 cutoff with a
        maximum length of lvl2_max_gap.

        lvl1_cutoff:  cutoff of value at enriched regions, default 500.
        lvl2_cutoff:  cutoff of value at linkage regions, default 100.
        min_length :  minimum peak length, default 200.
        lvl1_max_gap   :  maximum gap to merge nearby enriched peaks, default 50.
        lvl2_max_gap   :  maximum length of linkage regions, default 400.
        colname: can be 'sample','control','-100logp','-100logq'. Cutoff will be applied to the specified column.

        Return both general PeakIO object for highly enriched regions
        and gapped broad regions in BroadPeakIO.
        """
        cdef:
            bytes chrom
            int32_t i, j
            set chrs
            object lvl1, lvl2   # PeakContent class object
            list temppeakset, lvl1peakschrom, lvl2peakschrom
            
        assert lvl1_cutoff > lvl2_cutoff, "level 1 cutoff should be larger than level 2."
        assert lvl1_max_gap < lvl2_max_gap, "level 2 maximum gap should be larger than level 1."
        lvl1_peaks = self.call_peaks( cutoff=lvl1_cutoff, min_length=min_length, max_gap=lvl1_max_gap, call_summits=False )
        lvl2_peaks = self.call_peaks( cutoff=lvl2_cutoff, min_length=min_length, max_gap=lvl2_max_gap, call_summits=False )
        chrs = lvl1_peaks.get_chr_names()
        broadpeaks = BroadPeakIO()
        # use lvl2_peaks as linking regions between lvl1_peaks
        for chrom in sorted(chrs):
            lvl1peakschrom = lvl1_peaks.get_data_from_chrom(chrom)
            lvl2peakschrom = lvl2_peaks.get_data_from_chrom(chrom)
            lvl1peakschrom_next = iter(lvl1peakschrom).__next__
            tmppeakset = []             # to temporarily store lvl1 region inside a lvl2 region
            # our assumption is lvl1 regions should be included in lvl2 regions
            try:
                lvl1 = lvl1peakschrom_next()
                for i in range( len(lvl2peakschrom) ):
                    # for each lvl2 peak, find all lvl1 peaks inside
                    lvl2 = lvl2peakschrom[i]
                    while True:
                        if lvl2["start"] <= lvl1["start"]  and lvl1["end"] <= lvl2["end"]:
                            tmppeakset.append(lvl1)
                            lvl1 = lvl1peakschrom_next()
                        else:
                            self.__add_broadpeak ( broadpeaks, chrom, lvl2, tmppeakset)
                            tmppeakset = []
                            break
            except StopIteration:
                self.__add_broadpeak ( broadpeaks, chrom, lvl2, tmppeakset)
                tmppeakset = []
                for j in range( i+1, len(lvl2peakschrom) ):
                    self.__add_broadpeak ( broadpeaks, chrom, lvl2peakschrom[j], tmppeakset)
        return broadpeaks

    cdef object __add_broadpeak (self, object bpeaks, bytes chrom, object lvl2peak, list lvl1peakset):
        """Internal function to create broad peak.

        """
        cdef:
            int32_t start, end, blockNum
            bytes blockSizes, blockStarts, thickStart, thickEnd

        start      = lvl2peak["start"]
        end        = lvl2peak["end"]

        # the following code will add those broad/lvl2 peaks with no strong/lvl1 peaks inside
        if not lvl1peakset:
            # try:
            # will complement by adding 1bps start and end to this region
            # may change in the future if gappedPeak format was improved.
            bpeaks.add(chrom, start, end, score=lvl2peak["score"], thickStart=(b"%d" % start), thickEnd=(b"%d" % end),
                       blockNum = 2, blockSizes = b"1,1", blockStarts = (b"0,%d" % (end-start-1)), pileup = lvl2peak["pileup"],
                       pscore = lvl2peak["pscore"], fold_change = lvl2peak["fc"],
                       qscore = lvl2peak["qscore"] )
            return bpeaks

        thickStart = b"%d" % lvl1peakset[0]["start"]
        thickEnd   = b"%d" % lvl1peakset[-1]["end"]
        blockNum   = len(lvl1peakset)
        blockSizes = b",".join( [b"%d" % x["length"] for x in lvl1peakset] )
        blockStarts = b",".join( [b"%d" % (x["start"]-start) for x in lvl1peakset] )

        if int(thickStart) != start:
            # add 1bp left block
            thickStart = b"%d" % start
            blockNum += 1
            blockSizes = b"1,"+blockSizes
            blockStarts = b"0,"+blockStarts
        if int(thickEnd) != end:
            # add 1bp right block
            thickEnd = b"%d" % end
            blockNum += 1
            blockSizes = blockSizes+b",1"
            blockStarts = blockStarts + b"," + (b"%d" % (end-start-1))

        bpeaks.add(chrom, start, end, score=lvl2peak["score"], thickStart=thickStart, thickEnd=thickEnd,
                   blockNum = blockNum, blockSizes = blockSizes, blockStarts = blockStarts,  pileup = lvl2peak["pileup"],
                   pscore = lvl2peak["pscore"], fold_change = lvl2peak["fc"],
                   qscore = lvl2peak["qscore"] )
        return bpeaks


    cpdef int32_t total (self):
        """Return the number of regions in this object.

        """
        cdef:
            int32_t t
        t = 0
        for ( p, v ) in self.__data.values():
            t += len(p)
        return t

    cpdef object set_single_value (self, float32_t new_value):
        """Change all the values in bedGraph to the same new_value,
        return a new bedGraphTrackI.

        """
        cdef:
            bytes chrom
            int32_t max_p
            object ret

        ret = bedGraphTrackI()
        chroms = set(self.get_chr_names())
        for chrom in sorted(chroms):
            (p1,v1) = self.get_data_by_chr(chrom) # arrays for position and values
            # maximum p
            max_p = max(p1)
            # add a region from 0 to max_p
            ret.add_loc(chrom,0,max_p,new_value)
        return ret

    cpdef object overlie (self, object bdgTracks, str func="max" ):
        """Calculate two or more bedGraphTrackI objects by letting self
        overlying bdgTrack2, with user-defined functions.

        Transition positions from both bedGraphTrackI objects will be
        considered and combined. For example:

           #1 bedGraph (self)   |      #2 bedGraph
        -----------------------------------------------
        chr1  0    100  0       | chr1    0    150  1
        chr1  100  200  3       | chr1    150  250  2
        chr1  200  300  4       | chr1    250  300  4

        these two bedGraphs will be combined to have five transition
        points: 100, 150, 200, 250, and 300. So in order to calculate
        two bedGraphs, I pair values within the following regions
        like:

        chr   s   e     (#1,#2)   applied_func_max
        -----------------------------------------------
        chr1  0   100   (0,1)     1
        chr1  100 150   (3,1)     3
        chr1  150 200   (3,2)     3
        chr1  200 250   (4,2)     4
        chr1  250 300   (4,4)     4

        Then the given 'func' will be applied on each 2-tuple as func(#1,#2)

        Supported 'func' are "sum", "subtract" (only for two bdg
        objects), "product", "divide" (only for two bdg objects),
        "max", "mean" and "fisher".

        Return value is a new bedGraphTrackI object.

        Option: bdgTracks can be a list of bedGraphTrackI objects
        """
        cdef:
            int32_t pre_p, p1, p2
            float32_t v1, v2
            bytes chrom

        nr_tracks = len(bdgTracks) + 1  # +1 for self
        assert nr_tracks >= 2, "Specify at least one more bdg objects."
        for i, bdgTrack in enumerate(bdgTracks):
            assert isinstance(bdgTrack, bedGraphTrackI), "bdgTrack{} is not a bedGraphTrackI object".format(i + 1)

        if func == "max":
            f = max
        elif func == "mean":
            f = mean_func
        elif func == "fisher":
            f = fisher_func
        elif func == "sum":
            f = sum
        elif func == "product":
            f = product_func
        elif func == "subtract":
            if nr_tracks == 2:
                f = subtract_func
            else:
                raise Exception(f"Only one more bdg object is allowed, but provided {nr_tracks-1}")
        elif func == "divide":
            if nr_tracks == 2:
                f = divide_func
            else:
                raise Exception(f"Only one more bdg object is allowed, but provided {nr_tracks-1}")
        else:
            raise Exception("Invalid function {func}! Choose from 'sum', 'subtract' (only for two bdg objects), 'product', 'divide' (only for two bdg objects), 'max', 'mean' and 'fisher'. ")

        ret = bedGraphTrackI()
        retadd = ret.add_loc

        common_chr = set(self.get_chr_names())
        for track in bdgTracks:
            common_chr = common_chr.intersection(set(track.get_chr_names()))

        for chrom in sorted(common_chr):
            datas = [self.get_data_by_chr(chrom)]
            datas.extend([bdgTracks[i].get_data_by_chr(chrom) for i in range(len(bdgTracks))])

            ps, vs, pn, vn = [], [], [], []
            for data in datas:
                ps.append(data[0])
                pn.append(iter(ps[-1]).__next__)
                vs.append(data[1])
                vn.append(iter(vs[-1]).__next__)

            pre_p = 0                   # remember the previous position in the new bedGraphTrackI object ret
            try:
                ps_cur = [pn[i]() for i in range(len(pn))]
                vs_cur = [vn[i]() for i in range(len(pn))]

                while True:
                    # get the lowest position
                    lowest_p = min(ps_cur)

                    # at least one lowest position, could be multiple
                    locations = [i for i in range(len(ps_cur)) if ps_cur[i] == lowest_p]

                    # add the data until the interval
                    ret.add_loc(chrom, pre_p, ps_cur[locations[0]], f(vs_cur))

                    pre_p = ps_cur[locations[0]]
                    for index in locations:
                        ps_cur[index] = pn[index]()
                        vs_cur[index] = vn[index]()
            except StopIteration:
                # meet the end of either bedGraphTrackI, simply exit
                pass
        return ret

    cpdef bool apply_func ( self, func ):
        """Apply function 'func' to every value in this bedGraphTrackI object.

        *Two adjacent regions with same value after applying func will
        not be merged.
        """
        cdef int32_t i

        for (p,s) in self.__data.values():
            for i in range(len(s)):
                s[i] = func(s[i])
        self.maxvalue = func(self.maxvalue)
        self.minvalue = func(self.minvalue)
        return True

    cpdef p2q ( self ):
        """Convert pvalue scores to qvalue scores.

        *Assume scores in this bedGraph are pvalue scores! Not work
         for other type of scores.
        """
        cdef:
            bytes chrom
            object pos_array, pscore_array
            dict pvalue_stat = {}
            dict pqtable = {}
            int64_t n, pre_p, this_p, length, j, pre_l, l, i
            float32_t this_v, pre_v, v, q, pre_q, this_t, this_c
            int64_t N, k, this_l
            float32_t f
            int64_t nhcal = 0
            int64_t npcal = 0
            list unique_values
            float32_t t0, t1, t

        # calculate frequencies of each p-score
        for chrom in sorted(self.get_chr_names()):
            pre_p = 0

            [pos_array, pscore_array] = self.__data[ chrom ]

            pn = iter(pos_array).__next__
            vn = iter(pscore_array).__next__

            for i in range( len( pos_array ) ):
                this_p = pn()
                this_v = vn()
                this_l = this_p - pre_p
                if this_v in pvalue_stat:
                    pvalue_stat[ this_v ] += this_l
                else:
                    pvalue_stat[ this_v ] = this_l
                pre_p = this_p

            nhcal += len( pos_array )

        nhval = 0

        N = sum(pvalue_stat.values()) # total length
        k = 1                           # rank
        f = -log10(N)
        pre_v = -2147483647
        pre_l = 0
        pre_q = 2147483647              # save the previous q-value

        # calculate qscore for each pscore
        pqtable = {}
        unique_values = sorted(pvalue_stat.keys(), reverse=True)
        for i in range(len(unique_values)):
            v = unique_values[i]
            l = pvalue_stat[v]
            q = v + (log10(k) + f)
            q = max(0,min(pre_q,q))           # make q-score monotonic
            pqtable[ v ] = q
            pre_v = v
            pre_q = q
            k+=l
            nhcal += 1

        # convert pscore to qscore
        for chrom in sorted(self.get_chr_names()):
            [pos_array, pscore_array] = self.__data[ chrom ]

            for i in range( len( pos_array ) ):
                pscore_array[ i ] = pqtable[ pscore_array[ i ] ]

        self.merge_regions()
        return


    cpdef object extract_value ( self, object bdgTrack2 ):
        """Extract values from regions defined in bedGraphTrackI class object
        `regions`.

        """
        cdef:
            int32_t pre_p, p1, p2, i
            float32_t v1, v2
            bytes chrom
            object ret

        assert isinstance(bdgTrack2,bedGraphTrackI), "not a bedGraphTrackI object"

        ret = [ [], pyarray('f',[]), pyarray('L',[]) ] # 1: region in bdgTrack2; 2: value; 3: length with the value
        radd = ret[0].append
        vadd = ret[1].append
        ladd = ret[2].append

        chr1 = set(self.get_chr_names())
        chr2 = set(bdgTrack2.get_chr_names())
        common_chr = chr1.intersection(chr2)
        for i in range( len( common_chr ) ):
            chrom = common_chr.pop()
            (p1s,v1s) = self.get_data_by_chr(chrom) # arrays for position and values
            p1n = iter(p1s).__next__         # assign the next function to a viable to speed up
            v1n = iter(v1s).__next__

            (p2s,v2s) = bdgTrack2.get_data_by_chr(chrom) # arrays for position and values
            p2n = iter(p2s).__next__         # assign the next function to a viable to speed up
            v2n = iter(v2s).__next__
            pre_p = 0                   # remember the previous position in the new bedGraphTrackI object ret
            try:
                p1 = p1n()
                v1 = v1n()

                p2 = p2n()
                v2 = v2n()

                while True:
                    if p1 < p2:
                        # clip a region from pre_p to p1, then set pre_p as p1.
                        if v2>0:
                            radd(str(chrom)+"."+str(pre_p)+"."+str(p1))
                            vadd(v1)
                            ladd(p1-pre_p)
                        pre_p = p1
                        # call for the next p1 and v1
                        p1 = p1n()
                        v1 = v1n()
                    elif p2 < p1:
                        # clip a region from pre_p to p2, then set pre_p as p2.
                        if v2>0:
                            radd(str(chrom)+"."+str(pre_p)+"."+str(p2))
                            vadd(v1)
                            ladd(p2-pre_p)
                        pre_p = p2
                        # call for the next p2 and v2
                        p2 = p2n()
                        v2 = v2n()
                    elif p1 == p2:
                        # from pre_p to p1 or p2, then set pre_p as p1 or p2.
                        if v2>0:
                            radd(str(chrom)+"."+str(pre_p)+"."+str(p1))
                            vadd(v1)
                            ladd(p1-pre_p)
                        pre_p = p1
                        # call for the next p1, v1, p2, v2.
                        p1 = p1n()
                        v1 = v1n()
                        p2 = p2n()
                        v2 = v2n()
            except StopIteration:
                # meet the end of either bedGraphTrackI, simply exit
                pass

        return ret

    cpdef object extract_value_hmmr ( self, object bdgTrack2 ):
        """Extract values from regions defined in bedGraphTrackI class object
        `regions`.

        I will try to tweak this function to output only the values of
        bdgTrack1 (self) in the regions in bdgTrack2

        This is specifically for HMMRATAC. bdgTrack2 should be a
        bedgraph object containing the bins with value set to
        'mark_bin' -- the bins in the same region will have the same
        value.
        """
        cdef:
             int32_t pre_p, p1, p2, i
             float32_t v1, v2
             bytes chrom
             list ret

        assert isinstance(bdgTrack2,bedGraphTrackI), "not a bedGraphTrackI object"

        ret = [ [], pyarray('f',[]), pyarray('i',[]) ] # 0: bin location (chrom, position); 1: value; 2: number of bins in this region
        padd = ret[0].append
        vadd = ret[1].append
        ladd = ret[2].append

        chr1 = set(self.get_chr_names())
        chr2 = set(bdgTrack2.get_chr_names())
        common_chr = sorted(list(chr1.intersection(chr2)))
        for i in range( len( common_chr ) ):
            chrom = common_chr.pop()
            (p1s,v1s) = self.get_data_by_chr(chrom) # arrays for position and values
            p1n = iter(p1s).__next__         # assign the next function to a viable to speed up
            v1n = iter(v1s).__next__

            (p2s,v2s) = bdgTrack2.get_data_by_chr(chrom) # arrays for position and values
            p2n = iter(p2s).__next__         # assign the next function to a viable to speed up
            v2n = iter(v2s).__next__
            pre_p = 0                   # remember the previous position in the new bedGraphTrackI object ret
            try:
                p1 = p1n()
                v1 = v1n()

                p2 = p2n()
                v2 = v2n()

                while True:
                    if p1 < p2:
                        # clip a region from pre_p to p1, then set pre_p as p1.
                        # in this case, we don't output any
                        #if v2>0:
                        #    radd(str(chrom)+"."+str(pre_p)+"."+str(p1))
                        #    vadd(v1)
                        #    ladd(p1-pre_p)
                        pre_p = p1
                        # call for the next p1 and v1
                        p1 = p1n()
                        v1 = v1n()
                    elif p2 < p1:
                        # clip a region from pre_p to p2, then set pre_p as p2.
                        if v2 != 0: #0 means it's a gap region, we should have value > 1
                            padd( (chrom, p2) )
                            vadd(v1)
                            ladd(int(v2))
                        pre_p = p2
                        # call for the next p2 and v2
                        p2 = p2n()
                        v2 = v2n()
                    elif p1 == p2:
                        # from pre_p to p1 or p2, then set pre_p as p1 or p2.
                        if v2 != 0: #0 means it's a gap region, we should have 1 or -1
                            padd( (chrom, p2) )
                            vadd(v1)
                            ladd(int(v2))
                        pre_p = p1
                        # call for the next p1, v1, p2, v2.
                        p1 = p1n()
                        v1 = v1n()
                        p2 = p2n()
                        v2 = v2n()
            except StopIteration:
                # meet the end of either bedGraphTrackI, simply exit
                pass

        return ret

    cpdef make_ScoreTrackII_for_macs (self, object bdgTrack2, float32_t depth1 = 1.0, float32_t depth2 = 1.0 ):
        """A modified overlie function for MACS v2.

        effective_depth_in_million: sequencing depth in million after
                                    duplicates being filtered. If
                                    treatment is scaled down to
                                    control sample size, then this
                                    should be control sample size in
                                    million. And vice versa.

        Return value is a ScoreTrackII object.
        """
        cdef:
            int32_t pre_p, p1, p2
            float32_t v1, v2
            bytes chrom
            object ret

        assert isinstance(bdgTrack2,bedGraphTrackI), "bdgTrack2 is not a bedGraphTrackI object"

        ret = ScoreTrackII( treat_depth = depth1, ctrl_depth = depth2 )
        retadd = ret.add

        chr1 = set(self.get_chr_names())
        chr2 = set(bdgTrack2.get_chr_names())
        common_chr = chr1.intersection(chr2)
        for chrom in sorted(common_chr):

            (p1s,v1s) = self.get_data_by_chr(chrom) # arrays for position and values
            p1n = iter(p1s).__next__         # assign the next function to a viable to speed up
            v1n = iter(v1s).__next__

            (p2s,v2s) = bdgTrack2.get_data_by_chr(chrom) # arrays for position and values
            p2n = iter(p2s).__next__         # assign the next function to a viable to speed up
            v2n = iter(v2s).__next__

            chrom_max_len = len(p1s)+len(p2s) # this is the maximum number of locations needed to be recorded in scoreTrackI for this chromosome.

            ret.add_chromosome(chrom,chrom_max_len)

            pre_p = 0                   # remember the previous position in the new bedGraphTrackI object ret

            try:
                p1 = p1n()
                v1 = v1n()

                p2 = p2n()
                v2 = v2n()

                while True:
                    if p1 < p2:
                        # clip a region from pre_p to p1, then set pre_p as p1.
                        retadd( chrom, p1, v1, v2 )
                        pre_p = p1
                        # call for the next p1 and v1
                        p1 = p1n()
                        v1 = v1n()
                    elif p2 < p1:
                        # clip a region from pre_p to p2, then set pre_p as p2.
                        retadd( chrom, p2, v1, v2 )
                        pre_p = p2
                        # call for the next p2 and v2
                        p2 = p2n()
                        v2 = v2n()
                    elif p1 == p2:
                        # from pre_p to p1 or p2, then set pre_p as p1 or p2.
                        retadd( chrom, p1, v1, v2 )
                        pre_p = p1
                        # call for the next p1, v1, p2, v2.
                        p1 = p1n()
                        v1 = v1n()
                        p2 = p2n()
                        v2 = v2n()
            except StopIteration:
                # meet the end of either bedGraphTrackI, simply exit
                pass

        ret.finalize()
        #ret.merge_regions()
        return ret

    cpdef str cutoff_analysis ( self, int32_t max_gap, int32_t min_length, int32_t steps = 100, float32_t min_score = 0, float32_t max_score = 1000 ):
        """
        Cutoff analysis function for bedGraphTrackI object.
    
        This function will try all possible cutoff values on the score
        column to call peaks. Then will give a report of a number of
        metrics (number of peaks, total length of peaks, average
        length of peak) at varying score cutoffs. For each score
        cutoff, the function finds the positions where the score
        exceeds the cutoff, then groups those positions into "peaks"
        based on the maximum allowed gap (max_gap) between consecutive
        positions. If a peak's length exceeds the minimum length
        (min_length), the peak is counted.

        Parameters
        ----------

        max_gap : int32_t
        Maximum allowed gap between consecutive positions above cutoff
        
        min_length : int32_t Minimum length of peak
        steps: int32_t
        It will be used to calculate 'step' to increase from min_v to
        max_v (see below).

        min_score: float32_t
        Minimum score for cutoff analysis. Note1: we will take the
        larger value between the actual minimum value in the BedGraph
        and min_score as min_v. Note2: the min_v won't be included in
        the final result. We will try to output the smallest cutoff as
        min_v+step.

        max_score: float32_t
        Maximum score for cutoff analysis. Note1: we will take the
        smaller value between the actual maximum value in the BedGraph
        and max_score as max_v. Note2: the max_v may not be included
        in the final result. We will only output the cutoff that can
        generate at least 1 peak.

        Returns
        -------

        Cutoff analysis report in str object.

        Todos
        -----

        May need to seperate this function out as a class so that we
        can add more ways to analyze the result. Also, we can let this
        function return a list of dictionary or data.frame in that
        way, instead of str object.
        
        """
        cdef:
            set chrs
            list peak_content, ret_list, cutoff_list, cutoff_npeaks, cutoff_lpeaks
            bytes  chrom
            str ret
            float32_t cutoff
            int64_t total_l, total_p, i, n, ts, te, lastp, tl, peak_length
            #dict cutoff_npeaks, cutoff_lpeaks
            float32_t s, midvalue

        chrs = self.get_chr_names()

        #midvalue = self.minvalue/2 + self.maxvalue/2
        #s = float(self.minvalue - midvalue)/steps
        minv = max( min_score, self.minvalue )
        maxv = min( self.maxvalue, max_score )

        s = float(maxv - minv)/steps

        # a list of possible cutoff values from minv to maxv with step of s
        cutoff_list = [round(value, 3) for value in np.arange(minv, maxv, s)]

        cutoff_npeaks = [0] * len( cutoff_list )
        cutoff_lpeaks = [0] * len( cutoff_list )

        for chrom in sorted(chrs):
            ( pos_array, score_array ) = self.__data[ chrom ]
            pos_array = np.array( self.__data[ chrom ][ 0 ] )
            score_array = np.array( self.__data[ chrom ][ 1 ] )

            for n in range( len( cutoff_list ) ):
                cutoff = cutoff_list[ n ]
                total_l = 0           # total length of peaks
                total_p = 0           # total number of peaks

                # get the regions with scores above cutoffs
                above_cutoff = np.nonzero( score_array > cutoff )[0]# this is not an optimized method. It would be better to store score array in a 2-D ndarray?
                above_cutoff_endpos = pos_array[above_cutoff] # end positions of regions where score is above cutoff
                above_cutoff_startpos = pos_array[above_cutoff-1] # start positions of regions where score is above cutoff

                if above_cutoff_endpos.size == 0:
                    continue

                # first bit of region above cutoff
                acs_next = iter(above_cutoff_startpos).__next__
                ace_next = iter(above_cutoff_endpos).__next__

                ts = acs_next()
                te = ace_next()
                peak_content = [( ts, te ), ]
                lastp = te

                for i in range( 1, above_cutoff_startpos.size ):
                    ts = acs_next()
                    te = ace_next()
                    tl = ts - lastp
                    if tl <= max_gap:
                        peak_content.append( ( ts, te ) )
                    else:
                        peak_length = peak_content[ -1 ][ 1 ] - peak_content[ 0 ][ 0 ]
                        if peak_length >= min_length: # if the peak is too small, reject it
                            total_l +=  peak_length
                            total_p += 1
                        peak_content = [ ( ts, te ), ]
                    lastp = te

                if peak_content:
                    peak_length = peak_content[ -1 ][ 1 ] - peak_content[ 0 ][ 0 ]
                    if peak_length >= min_length: # if the peak is too small, reject it
                        total_l +=  peak_length
                        total_p += 1
                cutoff_lpeaks[ n ] += total_l
                cutoff_npeaks[ n ] += total_p
                
        # prepare the returnning text
        ret_list = ["score\tnpeaks\tlpeaks\tavelpeak\n"]
        for n in range( len( cutoff_list )-1, -1, -1 ):
            cutoff = cutoff_list[ n ]
            if cutoff_npeaks[ n ] > 0:
                ret_list.append("%.2f\t%d\t%d\t%.2f\n" % ( cutoff, cutoff_npeaks[ n ], \
                                                          cutoff_lpeaks[ n ], \
                                                          cutoff_lpeaks[ n ]/cutoff_npeaks[ n ] ))
        ret = ''.join(ret_list)
        return ret


cdef np.ndarray calculate_elbows( np.ndarray values, float32_t threshold=0.01):
    # although this function is supposed to find elbow pts for cutoff analysis, 
    # however, in reality, it barely works...
    cdef: 
        np.ndarray deltas, slopes, delta_slopes, elbows
        np.float32_t avg_delta_slope
        
    # Calculate the difference between each point and the first point
    deltas = values - values[0]
    
    # Calculate the slope between each point and the last point
    slopes = deltas / (values[-1] - values[0])
    
    # Calculate the change in slope
    delta_slopes = np.diff(slopes)
    
    # Calculate the average change in slope
    avg_delta_slope = np.mean(delta_slopes)
    
    # Find all points where the change in slope is significantly larger than the average
    elbows = np.where(delta_slopes > avg_delta_slope + threshold)[0]
    
    return elbows



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2020-12-04 23:10:35 Tao Liu>

"""Module for SAPPER PosReadsInfo class.

Copyright (c) 2017 Tao Liu <tliu4@buffalo.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: tliu4@buffalo.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
from MACS3.Signal.VariantStat import CalModel_Homo, CalModel_Heter_noAS, CalModel_Heter_AS, calculate_GQ, calculate_GQ_heterASsig
from MACS3.Signal.Prob import binomial_cdf
from MACS3.Signal.PeakVariants import Variant

from cpython cimport bool

import numpy as np
cimport numpy as np
from numpy cimport uint32_t, uint64_t, int32_t, float32_t

LN10 = 2.3025850929940458

cdef extern from "stdlib.h":
    ctypedef unsigned int size_t
    size_t strlen(char *s)
    void *malloc(size_t size)
    void *calloc(size_t n, size_t size)
    void free(void *ptr)
    int strcmp(char *a, char *b)
    char * strcpy(char *a, char *b)
    long atol(char *bytes)
    int atoi(char *bytes)

# ------------------------------------
# constants
# ------------------------------------
__version__ = "Parser $Revision$"
__author__ = "Tao Liu <tliu4@buffalo.edu>"
__doc__ = "All Parser classes"

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------

cdef class PosReadsInfo:
    cdef:
        long ref_pos
        bytes ref_allele
        bytes alt_allele
        bool filterout          # if true, do not output

        dict bq_set_T     #{A:[], C:[], G:[], T:[], N:[]} for treatment
        dict bq_set_C
        dict n_reads_T    #{A:[], C:[], G:[], T:[], N:[]} for treatment
        dict n_reads_C
        dict n_reads

        list n_strand #[{A:[], C:[], G:[], T:[], N:[]},{A:[], C:[], G:[], T:[], N:[]}] for total appearance on plus strand and minus strand for ChIP sample only
        dict n_tips # count of nt appearing at tips

        bytes top1allele
        bytes top2allele
        float top12alleles_ratio

        double lnL_homo_major,lnL_heter_AS,lnL_heter_noAS,lnL_homo_minor
        double BIC_homo_major,BIC_heter_AS,BIC_heter_noAS,BIC_homo_minor
        double PL_00, PL_01, PL_11
        double deltaBIC
        int heter_noAS_kc, heter_noAS_ki
        int heter_AS_kc, heter_AS_ki
        double heter_AS_alleleratio

        int GQ_homo_major,GQ_heter_noAS,GQ_heter_AS  #phred scale of prob by standard formular
        int GQ_heter_ASsig #phred scale of prob, to measure the difference between AS and noAS

        double GQ
        
        str GT
        str type
        str mutation_type       # SNV or Insertion or Deletion

        bool hasfermiinfor #if no fermi bam overlap in the position, false; if fermi bam in the position GT: N, false; if anyone of top2allele is not in fermi GT NTs, false;
        bytearray fermiNTs # 

    def __cinit__ ( self ):
        self.filterout = False
        self.GQ = 0
        self.GT = "unsure"
        self.alt_allele = b'.'
        
    def __init__ ( self, long ref_pos, bytes ref_allele ):
        self.ref_pos = ref_pos
        self.ref_allele = ref_allele
        self.bq_set_T = { ref_allele:[],b'A':[], b'C':[], b'G':[], b'T':[], b'N':[], b'*':[] }
        self.bq_set_C = { ref_allele:[],b'A':[], b'C':[], b'G':[], b'T':[], b'N':[], b'*':[] }
        self.n_reads_T = { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 }
        self.n_reads_C = { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 }
        self.n_reads =  { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 }
        self.n_strand = [ { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 }, { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 } ]
        self.n_tips = { ref_allele:0,b'A':0, b'C':0, b'G':0, b'T':0, b'N':0, b'*':0 }
        

    #cpdef void merge ( self, PosReadsInfo PRI2 ):
    #    """Merge two PRIs. No check available.
    #
    #    """
    #    assert self.ref_pos == PRI2.ref_pos
    #    assert self.ref_allele == PRI2.ref_allele
    #    for b in set( self.n_reads.keys() ).union( set( PRI2.n_reads.keys() ) ):
    #        self.bq_set_T[ b ] = self.bq_set_T.get( b, []).extend( PRI2.bq_set_T.get( b, [] ) )
    #        self.bq_set_C[ b ] = self.bq_set_C.get( b, []).extend( PRI2.bq_set_C.get( b, [] ) )
    #        self.n_reads_T[ b ] = self.n_reads_T.get( b, 0) + PRI2.n_reads_T.get( b, 0 )
    #        self.n_reads_C[ b ] = self.n_reads_C.get( b, 0) + PRI2.n_reads_C.get( b, 0 )
    #        self.n_reads[ b ] = self.n_reads.get( b, 0) + PRI2.n_reads.get( b, 0 )
    #    return

    def __getstate__ ( self ):
        return ( self.ref_pos, self.ref_allele, self.alt_allele, self.filterout,
                 self.bq_set_T, self.bq_set_C, self.n_reads_T, self.n_reads_C, self.n_reads, self.n_strand, self.n_tips,
                 self.top1allele, self.top2allele, self.top12alleles_ratio,
                 self.lnL_homo_major, self.lnL_heter_AS, self.lnL_heter_noAS, self.lnL_homo_minor,
                 self.BIC_homo_major, self.BIC_heter_AS, self.BIC_heter_noAS, self.BIC_homo_minor,
                 self.heter_noAS_kc, self.heter_noAS_ki,
                 self.heter_AS_kc, self.heter_AS_ki,
                 self.heter_AS_alleleratio,
                 self.GQ_homo_major, self.GQ_heter_noAS, self.GQ_heter_AS,
                 self.GQ_heter_ASsig,
                 self.GQ,
                 self.GT,
                 self.type,
                 self.hasfermiinfor,
                 self.fermiNTs )

    def __setstate__ ( self, state ):
        ( self.ref_pos, self.ref_allele, self.alt_allele, self.filterout,
          self.bq_set_T, self.bq_set_C, self.n_reads_T, self.n_reads_C, self.n_reads, self.n_strand, self.n_tips,
          self.top1allele, self.top2allele, self.top12alleles_ratio,
          self.lnL_homo_major, self.lnL_heter_AS, self.lnL_heter_noAS, self.lnL_homo_minor,
          self.BIC_homo_major, self.BIC_heter_AS, self.BIC_heter_noAS, self.BIC_homo_minor,
          self.heter_noAS_kc, self.heter_noAS_ki,
          self.heter_AS_kc, self.heter_AS_ki,
          self.heter_AS_alleleratio,
          self.GQ_homo_major, self.GQ_heter_noAS, self.GQ_heter_AS,
          self.GQ_heter_ASsig,
          self.GQ,
          self.GT,
          self.type,
          self.hasfermiinfor,
          self.fermiNTs ) = state

    cpdef bool filterflag ( self ):
        return self.filterout

    cpdef void apply_GQ_cutoff ( self, int min_homo_GQ = 50, int min_heter_GQ = 100 ):
        if self.filterout:
            return
        if self.type.startswith('homo') and self.GQ < min_homo_GQ:
            self.filterout = True
        elif self.type.startswith('heter') and self.GQ < min_heter_GQ:
            self.filterout = True
        return

    cpdef void apply_deltaBIC_cutoff ( self, float min_delta_BIC = 10 ):
        if self.filterout:
            return
        if self.deltaBIC < min_delta_BIC:
            self.filterout = True
        return

    cpdef void add_T ( self, int read_index, bytes read_allele, int read_bq, int strand, bool tip, int Q=20 ):
        """ Strand 0: plus, 1: minus

        Q is the quality cutoff. By default, only consider Q20 or read_bq > 20.
        """
        if read_bq <= Q:
            return
        if not self.n_reads.has_key( read_allele ):
            self.bq_set_T[read_allele] = []
            self.bq_set_C[read_allele] = []
            self.n_reads_T[read_allele] = 0
            self.n_reads_C[read_allele] = 0
            self.n_reads[read_allele] = 0
            self.n_strand[ 0 ][ read_allele ] = 0
            self.n_strand[ 1 ][ read_allele ] = 0
            self.n_tips[read_allele] = 0
        self.bq_set_T[read_allele].append( read_bq )
        self.n_reads_T[ read_allele ] += 1
        self.n_reads[ read_allele ] += 1
        self.n_strand[ strand ][ read_allele ] += 1
        if tip: self.n_tips[ read_allele ] += 1

    cpdef void add_C ( self, int read_index, bytes read_allele, int read_bq, int strand, int Q=20 ):
        if read_bq <= Q:
            return
        if not self.n_reads.has_key( read_allele ):
            self.bq_set_T[read_allele] = []
            self.bq_set_C[read_allele] = []
            self.n_reads_T[read_allele] = 0
            self.n_reads_C[read_allele] = 0
            self.n_reads[read_allele] = 0
            self.n_strand[ 0 ][ read_allele ] = 0
            self.n_strand[ 1 ][ read_allele ] = 0
            self.n_tips[read_allele] = 0
        self.bq_set_C[read_allele].append( read_bq )
        self.n_reads_C[ read_allele ] += 1
        self.n_reads[ read_allele ] += 1
        #self.n_strand[ strand ][ read_allele ] += 1

    cpdef int raw_read_depth ( self, str opt = "all" ):
        if opt == "all":
            return sum( self.n_reads.values() )
        elif opt == "T":
            return sum( self.n_reads_T.values() )
        elif opt == "C":
            return sum( self.n_reads_C.values() )
        else:
            raise Exception( "opt should be either 'all', 'T' or 'C'." )

    cpdef void update_top_alleles ( self, float min_top12alleles_ratio = 0.8, int min_altallele_count = 2, float max_allowed_ar = 0.95 ):
    #cpdef update_top_alleles ( self, float min_top12alleles_ratio = 0.8 ):
        """Identify top1 and top2 NT.  the ratio of (top1+top2)/total
        """
        cdef:
            float r

        [self.top1allele, self.top2allele] = sorted(self.n_reads, key=self.n_reads_T.get, reverse=True)[:2]

        # if top2 allele count in ChIP is lower than
        # min_altallele_count, or when allele ratio top1/(top1+top2)
        # is larger than max_allowed_ar in ChIP, we won't consider
        # this allele at all.  we set values of top2 allele in
        # dictionaries to [] and ignore top2 allele entirely.
 
        # max(self.n_strand[ 0 ][ self.top2allele ], self.n_strand[ 1 ][ self.top2allele ]) < min_altallele_count
        #if self.ref_pos == 52608504:
        #    print self.ref_pos, self.n_reads_T[ self.top1allele ], self.n_reads_T[ self.top2allele ], self.n_reads_C[ self.top1allele ], self.n_reads_C[ self.top2allele ]
        if self.n_reads_T[ self.top1allele ] + self.n_reads_T[ self.top2allele ] == 0:
            self.filterout = True
            return

        if (len(self.top1allele)==1 and len(self.top2allele)==1) and ( self.top2allele != self.ref_allele and ( ( self.n_reads_T[ self.top2allele ] - self.n_tips[ self.top2allele ] ) < min_altallele_count ) or \
                self.n_reads_T[ self.top1allele ]/(self.n_reads_T[ self.top1allele ] + self.n_reads_T[ self.top2allele ]) > max_allowed_ar ):
            self.bq_set_T[ self.top2allele ] = []
            self.bq_set_C[ self.top2allele ] = []
            self.n_reads_T[ self.top2allele ] = 0
            self.n_reads_C[ self.top2allele ] = 0
            self.n_reads[ self.top2allele ] = 0
            self.n_tips[ self.top2allele ] = 0
            if ( self.top1allele != self.ref_allele and ( self.n_reads_T[ self.top1allele ] - self.n_tips[ self.top1allele ] ) < min_altallele_count ):
                self.bq_set_T[ self.top1allele ] = []
                self.bq_set_C[ self.top1allele ] = []
                self.n_reads_T[ self.top1allele ] = 0
                self.n_reads_C[ self.top1allele ] = 0
                self.n_reads[ self.top1allele ] = 0
                self.n_tips[ self.top1allele ] = 0

        if self.n_reads_T[ self.top1allele ] + self.n_reads_T[ self.top2allele ] == 0:
            self.filterout = True
            return

        self.top12alleles_ratio = ( self.n_reads[ self.top1allele ] + self.n_reads[ self.top2allele ] ) /  sum( self.n_reads.values() )
        if self.top12alleles_ratio < min_top12alleles_ratio:
            self.filterout = True
            return

        if self.top1allele == self.ref_allele and self.n_reads[ self.top2allele ] == 0:
            # This means this position only contains top1allele which is the ref_allele. So the GT must be 0/0
            self.type = "homo_ref"
            self.filterout = True
            return
        return

    cpdef void top12alleles ( self ):
        print ( self.ref_pos, self.ref_allele)
        print ("Top1allele",self.top1allele, "Treatment", self.bq_set_T[self.top1allele], "Control", self.bq_set_C[self.top1allele])
        print ("Top2allele",self.top2allele, "Treatment", self.bq_set_T[self.top2allele], "Control", self.bq_set_C[self.top2allele])
    
    cpdef void call_GT ( self, float max_allowed_ar = 0.99 ):
        """Require update_top_alleles being called.
        """
        cdef:
            np.ndarray[np.int32_t, ndim=1] top1_bq_T
            np.ndarray[np.int32_t, ndim=1] top2_bq_T
            np.ndarray[np.int32_t, ndim=1] top1_bq_C
            np.ndarray[np.int32_t, ndim=1] top2_bq_C
            int i
            list top1_bq_T_l
            list top2_bq_T_l
            list top1_bq_C_l
            list top2_bq_C_l
            list tmp_mutation_type
            bytes tmp_alt

        if self.filterout:
            return
        
        top1_bq_T = np.array( self.bq_set_T[ self.top1allele ], dtype="int32" )
        top2_bq_T = np.array( self.bq_set_T[ self.top2allele ], dtype="int32" )
        top1_bq_C = np.array( self.bq_set_C[ self.top1allele ], dtype="int32" )
        top2_bq_C = np.array( self.bq_set_C[ self.top2allele ], dtype="int32" )
        (self.lnL_homo_major, self.BIC_homo_major) = CalModel_Homo( top1_bq_T, top1_bq_C, top2_bq_T, top2_bq_C )
        (self.lnL_homo_minor, self.BIC_homo_minor) = CalModel_Homo( top2_bq_T, top2_bq_C, top1_bq_T, top1_bq_C )
        (self.lnL_heter_noAS, self.BIC_heter_noAS) = CalModel_Heter_noAS( top1_bq_T, top1_bq_C, top2_bq_T, top2_bq_C )
        (self.lnL_heter_AS, self.BIC_heter_AS)     = CalModel_Heter_AS( top1_bq_T, top1_bq_C, top2_bq_T, top2_bq_C, max_allowed_ar )

        #if self.ref_pos == 71078525:
        #    print "---"
        #    print len( top1_bq_T ), len( top1_bq_C ), len( top2_bq_T ), len( top2_bq_C )
        #    print self.lnL_homo_major, self.lnL_homo_minor, self.lnL_heter_noAS, self.lnL_heter_AS
        #    print self.BIC_homo_major, self.BIC_homo_minor, self.BIC_heter_noAS, self.BIC_heter_AS
             
        if self.top1allele != self.ref_allele and self.n_reads[ self.top2allele ] == 0:
            # in this case, there is no top2 nt (or socalled minor
            # allele) in either treatment or control, we should assume
            # it's a 1/1 genotype. We will take 1/1 if it passes BIC
            # test (deltaBIC >=2), and will skip this loci if it can't
            # pass the test.
            
            self.deltaBIC = min( self.BIC_heter_noAS, self.BIC_heter_AS, self.BIC_homo_minor ) - self.BIC_homo_major
            if self.deltaBIC < 2:
               self.filterout = True
               return

            self.type = "homo"
            self.GT = "1/1"

            self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
            self.PL_01 = -10.0 * max( self.lnL_heter_noAS, self.lnL_heter_AS ) / LN10
            self.PL_11 = -10.0 * self.lnL_homo_major / LN10

            self.PL_00 = max( 0, self.PL_00 - self.PL_11 )
            self.PL_01 = max( 0, self.PL_01 - self.PL_11 )
            self.PL_11 = 0

            self.GQ = min( self.PL_00, self.PL_01 )
            self.alt_allele = self.top1allele
        else:
            # assign GQ, GT, and type
            if self.ref_allele != self.top1allele and self.BIC_homo_major + 2 <= self.BIC_homo_minor and self.BIC_homo_major + 2 <= self.BIC_heter_noAS and self.BIC_homo_major + 2 <= self.BIC_heter_AS:
                self.type = "homo"
                self.deltaBIC = min( self.BIC_heter_noAS, self.BIC_heter_AS, self.BIC_homo_minor ) - self.BIC_homo_major
                self.GT = "1/1"
                self.alt_allele = self.top1allele

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * max( self.lnL_heter_noAS, self.lnL_heter_AS ) / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_11
                self.PL_01 = self.PL_01 - self.PL_11
                self.PL_11 = 0

                self.GQ = min( self.PL_00, self.PL_01 )
                
            elif self.BIC_heter_noAS + 2 <= self.BIC_homo_major and self.BIC_heter_noAS + 2 <= self.BIC_homo_minor and self.BIC_heter_noAS + 2 <= self.BIC_heter_AS :
                self.type = "heter_noAS"
                self.deltaBIC = min( self.BIC_homo_major, self.BIC_homo_minor ) - self.BIC_heter_noAS

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * self.lnL_heter_noAS / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min( self.PL_00, self.PL_11 )
                
            elif self.BIC_heter_AS + 2 <= self.BIC_homo_major and self.BIC_heter_AS + 2 <= self.BIC_homo_minor and self.BIC_heter_AS + 2 <= self.BIC_heter_noAS:
                self.type = "heter_AS"
                self.deltaBIC = min( self.BIC_homo_major, self.BIC_homo_minor ) - self.BIC_heter_AS

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * self.lnL_heter_AS / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min( self.PL_00, self.PL_11 )

            elif self.BIC_heter_AS + 2 <= self.BIC_homo_major and self.BIC_heter_AS + 2 <= self.BIC_homo_minor:
                # can't decide if it's noAS or AS
                self.type = "heter_unsure"
                self.deltaBIC = min( self.BIC_homo_major, self.BIC_homo_minor ) - max( self.BIC_heter_AS, self.BIC_heter_noAS )

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * max( self.lnL_heter_noAS, self.lnL_heter_AS ) / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min( self.PL_00, self.PL_11 )
                
            elif self.ref_allele == self.top1allele and self.BIC_homo_major < self.BIC_homo_minor and self.BIC_homo_major < self.BIC_heter_noAS and self.BIC_homo_major < self.BIC_heter_AS:
                self.type = "homo_ref"
                # we do not calculate GQ if type is homo_ref
                self.GT = "0/0"
                self.filterout = True
            else:
                self.type="unsure"
                self.filterout = True

            if self.type.startswith( "heter" ):
                if self.ref_allele == self.top1allele:
                    self.alt_allele = self.top2allele
                    self.GT = "0/1"
                elif self.ref_allele == self.top2allele:
                    self.alt_allele = self.top1allele
                    self.GT = "0/1"
                else:
                    self.alt_allele = self.top1allele+b','+self.top2allele
                    self.GT = "1/2"
                # strand bias filter, uncomment following if wish to debug
                # calculate SB score
                #print "calculate SB score for ", self.ref_pos, "a/b/c/d:", self.n_strand[ 0 ][ self.top1allele ], self.n_strand[ 0 ][ self.top2allele ], self.n_strand[ 1 ][ self.top1allele ], self.n_strand[ 1 ][ self.top2allele ]
                #SBscore = self.SB_score_ChIP( self.n_strand[ 0 ][ self.top1allele ], self.n_strand[ 0 ][ self.top2allele ], self.n_strand[ 1 ][ self.top1allele ], self.n_strand[ 1 ][ self.top2allele ] )
                #SBscore = 0
                #if SBscore >= 1:
                #    print "disgard variant at", self.ref_pos, "type", self.type
                #    self.filterout = True
                
                # if self.ref_allele == self.top1allele：
                #     self.n_strand[ 0 ][ self.top1allele ] + self.n_strand[ 1 ][ self.top1allele ]
                #     if and self.n_strand[ 0 ][ self.top2allele ] == 0 or self.n_strand[ 1 ][ self.top2allele ] == 0:
                #         self.filterout = True
                #         print self.ref_pos


            # self.deltaBIC = self.deltaBIC

        tmp_mutation_type = []
        for tmp_alt in self.alt_allele.split(b','):
            if tmp_alt == b'*':
                tmp_mutation_type.append( "Deletion" )
            elif len( tmp_alt ) > 1:
                tmp_mutation_type.append( "Insertion" )
            else:
                tmp_mutation_type.append( "SNV" )
        self.mutation_type = ",".join( tmp_mutation_type )
        return

    cdef float SB_score_ChIP( self, int a, int b, int c, int d ):
        """ calculate score for filtering variants with strange strand biases.

        a: top1/major allele plus strand
        b: top2/minor allele plus strand
        c: top1/major allele minus strand
        d: top2/minor allele minus strand

        Return a float value so that if this value >= 1, the variant will be filtered out.
        """
        cdef:
            float score
            double p
            double p1_l, p1_r
            double p2_l, p2_r
            double top2_sb, top1_sb

        if a+b == 0 or c+d == 0:
            # if major allele and minor allele both bias to the same strand, allow it
            return 0.0

        # Rule:
        # if there is bias in top2 allele then bias in top1 allele should not be significantly smaller than it.
        # or there is no significant bias (0.5) in top2 allele.
        
        #print a, b, c, d
        p1_l = binomial_cdf( a, (a+c), 0.5, lower=True )      # alternative: less than 0.5
        p1_r = binomial_cdf( c, (a+c), 0.5, lower=True )   #              greater than 0.5
        p2_l = binomial_cdf( b, (b+d), 0.5, lower=True )      # alternative: less than 0.5
        p2_r = binomial_cdf( d, (b+d), 0.5, lower=True )   #              greater than 0.5
        #print p1_l, p1_r, p2_l, p2_r

        if (p1_l < 0.05 and p2_r < 0.05) or (p1_r < 0.05 and p2_l < 0.05):
            # we reject loci where the significant biases are inconsistent between top1 and top2 alleles.
            return 1.0
        else:
            # if b<=2 and d=0 or b=0 and d<=2 -- highly possible FPs
            #if ( b<=2 and d==0 or b==0 and d<=2 ):
            #    return 1
            # can't decide
            return 0.0

    cdef float SB_score_ATAC( self, int a, int b, int c, int d ):
        """ calculate score for filtering variants with strange strand biases.

        ATAC-seq version

        a: top1/major allele plus strand
        b: top2/minor allele plus strand
        c: top1/major allele minus strand
        d: top2/minor allele minus strand

        Return a float value so that if this value >= 1, the variant will be filtered out.
        """
        cdef:
            float score
            double p
            double p1_l, p1_r
            double p2_l, p2_r
            double top2_sb, top1_sb

        if a+b == 0 or c+d == 0:
            # if major allele and minor allele both bias to the same strand, allow it
            return 0.0

        # Rule:
        # if there is bias in top2 allele then bias in top1 allele should not be significantly smaller than it.
        # or there is no significant bias (0.5) in top2 allele.
        
        #print a, b, c, d
        p1_l = binomial_cdf( a, (a+c), 0.5, lower=True )      # alternative: less than 0.5
        p1_r = binomial_cdf( c, (a+c), 0.5, lower=True )   #              greater than 0.5
        p2_l = binomial_cdf( b, (b+d), 0.5, lower=True )      # alternative: less than 0.5
        p2_r = binomial_cdf( d, (b+d), 0.5, lower=True )   #              greater than 0.5
        #print p1_l, p1_r, p2_l, p2_r

        if (p1_l < 0.05 and p2_r < 0.05) or (p1_r < 0.05 and p2_l < 0.05):
            # we reject loci where the significant biases are inconsistent between top1 and top2 alleles.
            return 1.0
        else:
            # can't decide
            return 0.0

    cpdef str to_vcf ( self ):
        """Output REF,ALT,QUAL,FILTER,INFO,FORMAT, SAMPLE columns.
        """
        cdef:
            str vcf_ref, vcf_alt, vcf_qual, vcf_filter, vcf_info, vcf_format, vcf_sample

        vcf_ref = self.ref_allele.decode()
        vcf_alt = self.alt_allele.decode()
        vcf_qual = "%d" % self.GQ
        vcf_filter = "."
        vcf_info = (b"M=%s;MT=%s;DPT=%d;DPC=%d;DP1T=%d%s;DP2T=%d%s;DP1C=%d%s;DP2C=%d%s;SB=%d,%d,%d,%d;DBIC=%.2f;BICHOMOMAJOR=%.2f;BICHOMOMINOR=%.2f;BICHETERNOAS=%.2f;BICHETERAS=%.2f;AR=%.2f" % \
            (self.type.encode(), self.mutation_type.encode(), sum( self.n_reads_T.values() ), sum( self.n_reads_C.values() ), 
             self.n_reads_T[self.top1allele], self.top1allele, self.n_reads_T[self.top2allele], self.top2allele,
             self.n_reads_C[self.top1allele], self.top1allele, self.n_reads_C[self.top2allele], self.top2allele,
             self.n_strand[ 0 ][ self.top1allele ], self.n_strand[ 0 ][ self.top2allele ], self.n_strand[ 1 ][ self.top1allele ], self.n_strand[ 1 ][ self.top2allele ],
             self.deltaBIC,
             self.BIC_homo_major, self.BIC_homo_minor, self.BIC_heter_noAS,self.BIC_heter_AS,
             self.n_reads_T[self.top1allele]/(self.n_reads_T[self.top1allele]+self.n_reads_T[self.top2allele])
             )).decode()
        vcf_format = "GT:DP:GQ:PL"
        vcf_sample = "%s:%d:%d:%d,%d,%d" % (self.GT, self.raw_read_depth( opt = "all" ), self.GQ, self.PL_00, self.PL_01, self.PL_11)
        return "\t".join( ( vcf_ref, vcf_alt, vcf_qual, vcf_filter, vcf_info, vcf_format, vcf_sample ) )

    cpdef toVariant ( self ):
        cdef:
            object v
        v = Variant( 
                     self.ref_allele.decode(),
                     self.alt_allele.decode(),
                     self.GQ,
                     '.',
                     self.type,
                     self.mutation_type,
                     self.top1allele.decode(),
                     self.top2allele.decode(),
                     sum( self.n_reads_T.values() ),
                     sum( self.n_reads_C.values() ),
                     self.n_reads_T[self.top1allele],
                     self.n_reads_T[self.top2allele],
                     self.n_reads_C[self.top1allele],
                     self.n_reads_C[self.top2allele],
                     self.n_strand[ 0 ][ self.top1allele ],
                     self.n_strand[ 0 ][ self.top2allele ],
                     self.n_strand[ 1 ][ self.top1allele ],
                     self.n_strand[ 1 ][ self.top2allele ],
                     self.deltaBIC,
                     self.BIC_homo_major,
                     self.BIC_homo_minor,
                     self.BIC_heter_noAS,
                     self.BIC_heter_AS,
                     self.n_reads_T[self.top1allele]/(self.n_reads_T[self.top1allele]+self.n_reads_T[self.top2allele]),
                     self.GT,
                     self.raw_read_depth( opt = "all" ),
                     self.PL_00,
                     self.PL_01,
                     self.PL_11 )
        return v
